# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Test conversion between NIFTI and NIPY conventions.  The algorithms are
mostly written out in the :mod:`nipy.io.nifti_ref` docstrings.
"""
import warnings

from copy import copy

import numpy as np

import nibabel as nib
from nibabel.affines import from_matvec
from nibabel.spatialimages import HeaderDataError

from ...core.api import (Image,
                         AffineTransform as AT,
                         CoordinateSystem as CS)
from ...core.reference.spaces import (unknown_csm, scanner_csm, aligned_csm,
                                      talairach_csm, mni_csm, unknown_space,
                                      vox2mni)

from ..files import load
from ..nifti_ref import (nipy2nifti, nifti2nipy, NiftiError)

from nose.tools import assert_equal, assert_true, assert_false, assert_raises
from numpy.testing import assert_almost_equal, assert_array_equal

from ...testing import anatfile, funcfile

def copy_of(fname):
    # Make a fresh copy of a image stored in a file
    img = load(fname)
    hdr = img.metadata['header'].copy()
    return Image(img.get_data().copy(),
                 copy(img.coordmap),
                 {'header': hdr})


def setup():
    # Suppress warnings during tests
    warnings.simplefilter("ignore")


def teardown():
    # Clear list of warning filters
    warnings.resetwarnings()


def test_basic_nipy2nifti():
    # Go from nipy image to header and data for nifti
    fimg = copy_of(funcfile)
    hdr = fimg.metadata['header']
    data = fimg.get_data()
    # Header is preserved
    # Put in some information to check header is preserved
    hdr['slice_duration'] = 0.25
    ni_img = nipy2nifti(fimg)
    new_hdr = ni_img.get_header()
    # header copied on the way through
    assert_false(hdr is new_hdr)
    # Check information preserved
    assert_equal(hdr['slice_duration'], new_hdr['slice_duration'])
    assert_array_equal(data, ni_img.get_data())
    # Shape obviously should be same
    assert_equal(ni_img.shape, fimg.shape)


def test_xyz_affines():
    fimg = copy_of(funcfile)
    data = fimg.get_data()
    # Check conversion to xyz affable
    # Roll time to front in array
    fimg_t0 = fimg.reordered_axes((3, 0, 1, 2))
    # Nifti conversion rolls it back
    assert_array_equal(nipy2nifti(fimg_t0).get_data(), data)
    # Roll time to position 1
    fimg_t0 = fimg.reordered_axes((0, 3, 1, 2))
    assert_array_equal(nipy2nifti(fimg_t0).get_data(), data)
    # Check bad names cause NiftiError
    out_coords = fimg.reference.coord_names
    bad_img = fimg.renamed_reference(**{out_coords[0]: 'not a known axis'})
    assert_raises(NiftiError, nipy2nifti, bad_img)
    # Check xyz works for not strict
    bad_img = fimg.renamed_reference(**dict(zip(out_coords, 'xyz')))
    assert_array_equal(nipy2nifti(bad_img, strict=False).get_data(), data)
    # But fails for strict
    assert_raises(NiftiError, nipy2nifti, bad_img, strict=True)
    # 3D is OK
    aimg = copy_of(anatfile)
    adata = aimg.get_data()
    assert_array_equal(nipy2nifti(aimg).get_data(), adata)
    # For now, always error on 2D (this depends on as_xyz_image)
    assert_raises(NiftiError, nipy2nifti, aimg[:, :, 1])
    assert_raises(NiftiError, nipy2nifti, aimg[:, 1, :])
    assert_raises(NiftiError, nipy2nifti, aimg[1, :, :])
    # Do not allow spaces not in the NIFTI canon
    for i in range(3):
        displaced_img = fimg.renamed_reference(**{out_coords[i]: 'obscure'})
        assert_raises(NiftiError, nipy2nifti, displaced_img)


def test_unknown():
    # The 'unknown' coordinate space results from loading an image with no
    # affine set; allow setting into nifti iff the affine corresponds to the
    # default that would be created when there is no affine
    aimg = copy_of(anatfile)
    bare_affine = aimg.metadata['header'].get_base_affine()
    # The affine does not match the header-only affine
    assert_false(np.allclose(bare_affine, aimg.coordmap.affine))
    unknown_cs = unknown_csm(3)
    out_coords = aimg.reference.coord_names
    # So nipy2nifti raises an error
    displaced_img = aimg.renamed_reference(
        **dict(zip(out_coords[:3], unknown_cs.coord_names)))
    assert_raises(NiftiError, nipy2nifti, displaced_img)
    # If the affine is the same, no error
    displaced_img.coordmap.affine[:] = bare_affine
    assert_true(np.allclose(bare_affine, displaced_img.coordmap.affine))
    nimg = nipy2nifti(displaced_img)
    assert_array_equal(nimg.get_affine(), bare_affine)
    inimg = nifti2nipy(nimg)
    assert_true(inimg.coordmap.function_range in unknown_space)


def test_orthogonal_dims():
    # Test whether conversion to nifti raises an error for non-orthogonal
    # non-spatial dimensions
    # This affine is all nicely diagonal
    aff = from_matvec(np.diag([2., 3, 4, 5, 6]), [10, 11, 12, 13, 14])
    data = np.random.normal(size=(3, 4, 5, 6, 7))
    img = Image(data, vox2mni(aff))
    def as3d(aff):
        return from_matvec(aff[:3, :3], aff[:3, -1])
    assert_array_equal(nipy2nifti(img).get_affine(), as3d(aff))
    # Non-orthogonal spatial dimensions OK
    aff[:3, :3] = np.random.normal(size=(3, 3))
    img = Image(data, vox2mni(aff))
    assert_array_equal(nipy2nifti(img).get_affine(), as3d(aff))
    # Space must be orthogonal to time etc
    aff[0, 3] = 0.1
    assert_raises(NiftiError, nipy2nifti, img)
    aff[0, 3] = 0
    assert_array_equal(nipy2nifti(img).get_affine(), as3d(aff))
    aff[3, 0] = 0.1
    assert_raises(NiftiError, nipy2nifti, img)
    aff[3, 0] = 0
    assert_array_equal(nipy2nifti(img).get_affine(), as3d(aff))
    aff[4, 0] = 0.1
    assert_raises(NiftiError, nipy2nifti, img)


def test_dim_info():
    # Test slice, freq, phase get set OK
    fimg = copy_of(funcfile)
    hdr = fimg.metadata['header']
    assert_equal(hdr.get_dim_info(), (None, None, None))
    ni_img = nipy2nifti(fimg)
    assert_equal(ni_img.get_header().get_dim_info(), (None, None, None))
    data = fimg.get_data()
    cmap = fimg.coordmap
    for i in range(3):
        for order, name in enumerate(('freq', 'phase', 'slice')):
            ncmap = cmap.renamed_domain({i: name})
            ni_img = nipy2nifti(Image(data, ncmap, {'header': hdr}))
            exp_info = [None, None, None]
            exp_info[order] = i
            assert_equal(ni_img.get_header().get_dim_info(),
                         tuple(exp_info))
    ncmap = cmap.renamed_domain(
        dict(zip(range(3), ('phase', 'slice', 'freq'))))
    ni_img = nipy2nifti(Image(data, ncmap, {'header': hdr}))
    assert_equal(ni_img.get_header().get_dim_info(), (2, 0, 1))


def test_time_like_matching():
    # Check checks for matching time-like axes
    shape = (2, 3, 4, 5, 6)
    shape_shifted = (2, 3, 4, 6, 5)
    data = np.random.normal(size=shape)
    aff = np.diag([3, 4, 5, 6, 7, 1])
    mni_names = mni_csm(3).coord_names
    time_cans = ('t', 'hz', 'ppm', 'rads')
    aliases = dict(t='time',
                   hz='frequency-hz',
                   ppm='concentration-ppm',
                   rads='radians/s')
    all_names = set(time_cans + tuple(v for v in aliases.values()))
    for time_like in time_cans:
        alias = aliases[time_like]
        for name in (time_like, alias):
            # Names match
            cmap = AT(CS(('i', 'j', 'k', name, 'u')),
                    CS(mni_names + (name, 'u')), aff)
            assert_equal(nipy2nifti(Image(data, cmap)).shape, shape)
            cmap = AT(CS(('i', 'j', 'k', 'u', name)),
                    CS(mni_names + ('u', name)), aff)
            assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)
            # No time-like in output is OK
            cmap = AT(CS(('i', 'j', 'k', 'u', name)),
                    CS(mni_names + ('u', 'v')), aff)
            assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)
            # No time-like in input is OK
            cmap = AT(CS(('i', 'j', 'k', 'u', 'v')),
                    CS(mni_names + ('u', name)), aff)
            assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)
            # Time-like in both, but not matching, not OK
            cmap = AT(CS(('i', 'j', 'k', 'u', name)),
                    CS(mni_names + (name, 'u')), aff)
            assert_raises(NiftiError, nipy2nifti, Image(data, cmap))
            # Time like in both with no match between but no match elsewhere
            # Actually this does cause a problem for non-zero time offset and
            # time axes, but we test that elsewhere.
            cmap = AT(CS(('i', 'j', 'k', 'u', name)),
                    CS(mni_names + ('u', name)),
                    np.diag([3, 4, 5, 6, 0, 1]))
            assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)
            cmap = AT(CS(('i', 'j', 'k', 'u', name)),
                    CS(mni_names + (name, 'u')),
                    np.diag([3, 4, 5, 0, 0, 1]))
            assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)
        # Matching to own alias is OK
        cmap = AT(CS(('i', 'j', 'k', time_like, 'u')),
                CS(mni_names + (alias, 'u')), aff)
        assert_equal(nipy2nifti(Image(data, cmap)).shape, shape)
        cmap = AT(CS(('i', 'j', 'k', alias, 'u')),
                CS(mni_names + (time_like, 'u')), aff)
        assert_equal(nipy2nifti(Image(data, cmap)).shape, shape)
        # But not to another time-like name
        others = all_names.difference((time_like, alias))
        for name in others:
            cmap = AT(CS(('i', 'j', 'k', time_like, 'u')),
                    CS(mni_names + (name, 'u')), aff)
            assert_raises(NiftiError, nipy2nifti, Image(data, cmap))
            cmap = AT(CS(('i', 'j', 'k', name, 'u')),
                    CS(mni_names + (time_like, 'u')), aff)
            assert_raises(NiftiError, nipy2nifti, Image(data, cmap))
        # It's OK to have more than one time-like, but the order of recognition
        # is 't', 'hz', 'ppm', 'rads'
        for i, better in enumerate(time_cans[:-1]):
            for worse in time_cans[i+1:]:
                cmap = AT(CS(('i', 'j', 'k', better, worse)),
                        CS(mni_names + (better, worse)), aff)
                assert_equal(nipy2nifti(Image(data, cmap)).shape, shape)
                cmap = AT(CS(('i', 'j', 'k', worse, better)),
                        CS(mni_names + (worse, better)), aff)
                assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)
                # Even if better is only in output
                cmap = AT(CS(('i', 'j', 'k', worse, 'u')),
                        CS(mni_names + (worse, better)), aff)
                assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)


def test_time_pixdims():
    # Pixdims get moved across when a no-time extra axis is added
    shape = (2, 3, 4, 5, 6, 7)
    data = np.random.normal(size=shape)
    aff = np.diag([3, 4, 5, 6, 7, 8, 1])
    mni_names = mni_csm(3).coord_names
    in_cs = CS('ikjlmn')
    cmap = AT(in_cs, CS(mni_names + tuple('tuv')), aff)
    hdr = nipy2nifti(Image(data, cmap)).get_header()
    assert_equal(hdr.get_zooms(), (3, 4, 5, 6, 7, 8))
    cmap = AT(in_cs, CS(mni_names + tuple('quv')), aff)
    hdr = nipy2nifti(Image(data, cmap)).get_header()
    assert_equal(hdr.get_zooms(), (3, 4, 5, 0, 6, 7, 8))


def test_xyzt_units():
    # Whether xyzt_unit field gets set correctly
    fimg_orig = copy_of(funcfile)
    # Put time in output, input and both
    data = fimg_orig.get_data()
    hdr = fimg_orig.metadata['header']
    aff = fimg_orig.coordmap.affine
    out_names = fimg_orig.reference.coord_names
    # Time in input only
    cmap_t_in = AT('ijkt', out_names[:3] + ('q',), aff)
    img_t_in = Image(data, cmap_t_in, {'header': hdr.copy()})
    # Time in output only
    cmap_t_out = AT('ijkl', out_names[:3] + ('t',), aff)
    img_t_out = Image(data, cmap_t_out, {'header': hdr.copy()})
    # Time in both
    cmap_t_b = AT('ijkt', out_names[:3] + ('t',), aff)
    img_t_b = Image(data, cmap_t_b, {'header': hdr.copy()})
    # In neither
    cmap_t_no = AT('ijkl', out_names[:3] + ('q',), aff)
    img_t_no = Image(data, cmap_t_no, {'header': hdr.copy()})
    # Check the default
    assert_equal(hdr.get_xyzt_units(), ('mm', 'sec'))
    # That default survives nifti conversion
    for img in (img_t_in, img_t_out, img_t_b):
        ni_img = nipy2nifti(img)
        assert_equal(ni_img.get_header().get_xyzt_units(), ('mm', 'sec'))
    # Now with no time
    for img in (img_t_no, img_t_b[...,0]):
        ni_img = nipy2nifti(img)
        assert_equal(ni_img.get_header().get_xyzt_units(), ('mm', 'unknown'))
    # Change to other time-like
    for units, name0, name1 in (('sec', 't', 'time'),
                                ('hz', 'hz', 'frequency-hz'),
                                ('ppm', 'ppm', 'concentration-ppm'),
                                ('rads', 'rads', 'radians/s')):
        for name in (name0, name1):
            new_img = img_t_out.renamed_reference(t=name)
            ni_img = nipy2nifti(new_img)
            assert_equal(ni_img.get_header().get_xyzt_units(), ('mm', units))
            new_img = img_t_in.renamed_axes(t=name)
            ni_img = nipy2nifti(new_img)
            assert_equal(ni_img.get_header().get_xyzt_units(), ('mm', units))
            new_img = img_t_b.renamed_axes(t=name).renamed_reference(t=name)
            ni_img = nipy2nifti(new_img)
            assert_equal(ni_img.get_header().get_xyzt_units(), ('mm', units))


def test_time_axes_4th():
    # Check time-like axes rolled to be 4th, and pixdims match
    data = np.random.normal(size=(2, 3, 4, 5, 6, 7))
    aff = np.diag([2., 3, 4, 5, 6, 7, 1])
    xyz_names = talairach_csm(3).coord_names
    in_cs = CS('ijklmn')
    for time_like in ('t', 'hz', 'ppm', 'rads'):
        cmap = AT(in_cs, CS(xyz_names + (time_like, 'q', 'r')), aff)
        img = Image(data, cmap)
        # Time-like in correct position
        ni_img = nipy2nifti(img)
        assert_array_equal(ni_img.get_data(), data)
        assert_array_equal(ni_img.get_header().get_zooms(), (2, 3, 4, 5, 6, 7))
        # Time-like needs reordering
        cmap = AT(in_cs, CS(xyz_names + ('q', time_like, 'r')), aff)
        ni_img = nipy2nifti(Image(data, cmap))
        assert_array_equal(ni_img.get_data(), np.rollaxis(data, 4, 3))
        assert_array_equal(ni_img.get_header().get_zooms(), (2, 3, 4, 6, 5, 7))
        # And again
        cmap = AT(in_cs, CS(xyz_names + ('q', 'r', time_like)), aff)
        ni_img = nipy2nifti(Image(data, cmap))
        assert_array_equal(ni_img.get_data(), np.rollaxis(data, 5, 3))
        assert_array_equal(ni_img.get_header().get_zooms(), (2, 3, 4, 7, 5, 6))


def test_save_toffset():
    # Check toffset only gets set for time
    shape = (2, 3, 4, 5, 6, 7)
    data = np.random.normal(size = shape)
    aff = from_matvec(np.diag([2., 3, 4, 5, 6, 7]),
                              [11, 12, 13, 14, 15, 16])
    xyz_names = talairach_csm(3).coord_names
    in_cs = CS('ijklmn')
    for t_name in 't', 'time':
        cmap = AT(in_cs, CS(xyz_names + (t_name, 'q', 'r')), aff)
        ni_img = nipy2nifti(Image(data, cmap))
        assert_equal(ni_img.get_header()['toffset'], 14)
    for time_like in ('hz', 'ppm', 'rads'):
        cmap = AT(in_cs, CS(xyz_names + (time_like, 'q', 'r')), aff)
        ni_img = nipy2nifti(Image(data, cmap))
        assert_equal(ni_img.get_header()['toffset'], 0)
    # Check that non-matching time causes a nifti error when toffset !=0
    shape_shifted = (2, 3, 4, 6, 5, 7)
    for t_name in 't', 'time':
        # No toffset, this is OK
        cmap = AT(CS(('i', 'j', 'k', 'u', t_name, 'v')),
                  CS(xyz_names + ('u', t_name, 'v')),
                  np.diag([3, 4, 5, 6, 0, 7, 1]))
        assert_equal(nipy2nifti(Image(data, cmap)).shape, shape_shifted)
        # toffset with 0 on TR (time) diagonal
        aff_z1 = from_matvec(np.diag([2., 3, 4, 5, 0, 7]),
                             [11, 12, 13, 14, 15, 16])
        cmap = AT(CS(('i', 'j', 'k', 'u', t_name, 'v')),
                  CS(xyz_names + ('u', t_name, 'v')),
                  aff_z1)
        # Default is to fix the zero
        assert_equal(nipy2nifti(Image(data, cmap)).shape,
                     shape_shifted)
        assert_equal(nipy2nifti(Image(data, cmap), fix0=True).shape,
                     shape_shifted)
        # Unless fix0 is False
        assert_raises(NiftiError, nipy2nifti, Image(data, cmap), fix0=False)
        # Fix doesn't work if there is more than one zero row and column
        aff_z2 = from_matvec(np.diag([2., 3, 4, 0, 0, 7]),
                             [11, 12, 13, 14, 15, 16])
        cmap = AT(CS(('i', 'j', 'k', 'u', t_name, 'v')),
                  CS(xyz_names + ('u', t_name, 'v')),
                  aff_z2)
        assert_raises(NiftiError, nipy2nifti, Image(data, cmap), fix0=True)
    # zeros on the diagonal are not a problem for non-time, with toffset,
    # because we don't need to set the 'time' part of the translation vector,
    # and therefore we don't need to know which *output axis* is time-like
    for t_name in 'hz', 'ppm', 'rads':
        cmap = AT(CS(('i', 'j', 'k', 'u', t_name, 'v')),
                  CS(xyz_names + ('u', t_name, 'v')),
                  aff_z1)
        assert_equal(nipy2nifti(Image(data, cmap), fix0=False).shape,
                     shape_shifted)
        cmap = AT(CS(('i', 'j', 'k', 'u', t_name, 'v')),
                  CS(xyz_names + ('u', t_name, 'v')),
                  aff_z2)
        assert_equal(nipy2nifti(Image(data, cmap), fix0=False).shape,
                     shape_shifted)


def test_too_many_dims():
    data0 = np.zeros(range(2, 9))
    xyz_names = talairach_csm(3).coord_names
    cmap = AT(CS('ijktuvw'), CS(xyz_names + tuple('tuvw')), np.eye(8))
    assert_equal(nipy2nifti(Image(data0, cmap)).shape, tuple(range(2, 9)))
    # Too many dimensions
    data1 = np.zeros(range(2, 10))
    cmap = AT(CS('ijktuvwq'), CS(xyz_names + tuple('tuvwq')), np.eye(9))
    assert_raises(NiftiError, nipy2nifti, Image(data1, cmap))
    # No time adds a dimension
    cmap = AT(CS('ijkpuvw'), CS(xyz_names + tuple('puvw')), np.eye(8))
    assert_raises(NiftiError, nipy2nifti, Image(data0, cmap))


def test_no_time():
    # Check that no time axis results in extra length 1 dimension
    data = np.random.normal(size=(2, 3, 4, 5, 6, 7))
    aff = np.diag([2., 3, 4, 5, 6, 7, 1])
    xyz_names = talairach_csm(3).coord_names
    in_cs = CS('ijklmn')
    # No change in shape if there's a time-like axis
    for time_like in ('t', 'hz', 'ppm', 'rads'):
        cmap = AT(in_cs, CS(xyz_names + (time_like, 'q', 'r')), aff)
        ni_img = nipy2nifti(Image(data, cmap))
        assert_array_equal(ni_img.get_data(), data)
    # But there is if no time-like
    for no_time in ('random', 'words', 'I', 'thought', 'of'):
        cmap = AT(in_cs, CS(xyz_names + (no_time, 'q', 'r')), aff)
        ni_img = nipy2nifti(Image(data, cmap))
        assert_array_equal(ni_img.get_data(), data[:, :, :, None, :, :])


def test_save_spaces():
    # Test that intended output spaces get set into nifti
    data = np.random.normal(size=(2, 3, 4))
    aff = np.diag([2., 3, 4, 1])
    in_cs = CS('ijk')
    for label, csm in (('scanner', scanner_csm),
                       ('aligned', aligned_csm),
                       ('talairach', talairach_csm),
                       ('mni', mni_csm)):
        img = Image(data, AT(in_cs, csm(3), aff))
        ni_img = nipy2nifti(img)
        assert_equal(ni_img.get_header().get_value_label('sform_code'),
                     label)


def test_save_dtype():
    # Test we can specify the dtype on conversion
    data = np.random.normal(size=(2, 3, 4))
    cmap = vox2mni(np.diag([2., 3, 4, 1]))
    for dt_code in ('i1', 'u1', 'i2', 'u2', 'i4', 'u4', 'i8', 'u8',
                    'f4', 'f8', 'c8', 'c16'):
        dt = np.dtype(dt_code)
        img = Image(data.astype(dt_code), cmap)
        ni_img = nipy2nifti(img, data_dtype=dt_code)
        assert_equal(ni_img.get_header().get_data_dtype(), dt)
        ni_img = nipy2nifti(img, data_dtype=dt)
        assert_equal(ni_img.get_header().get_data_dtype(), dt)
    # None results in trying to get the code from the input header, then from the
    # data.
    # From data, when there's nothing in the header
    img = Image(data.astype(np.int16), cmap)
    ni_img = nipy2nifti(img, data_dtype=None)
    assert_equal(ni_img.get_header().get_data_dtype(), np.dtype(np.int16))
    # From the header
    hdr = nib.Nifti1Header()
    hdr.set_data_dtype(np.int32)
    img = Image(data.astype(np.int16), cmap, metadata={'header': hdr})
    ni_img = nipy2nifti(img, data_dtype=None)
    assert_equal(ni_img.get_header().get_data_dtype(), np.dtype(np.int32))
    # Bad dtype
    assert_raises(TypeError, nipy2nifti, img, data_dtype='foo')
    # Fancy dtype
    data = np.zeros((2, 3, 4), dtype=[('f0', 'i2'), ('f1', 'f4')])
    img = Image(data, cmap)
    assert_raises(HeaderDataError, nipy2nifti, img, data_dtype=None)


def test_basic_load():
    # Just basic load
    data = np.random.normal(size=(2, 3, 4, 5))
    aff = np.diag([2., 3, 4, 1])
    ni_img = nib.Nifti1Image(data, aff)
    img = nifti2nipy(ni_img)
    assert_array_equal(img.get_data(), data)


def test_expand_to_3d():
    # Test 1D and 2D niftis
    # 1D and 2D with full sform or qform affines raise a NiftiError, because we
    # can't be sure which axes the affine refers to.  Should the image have 1
    # length axes prepended?  Or appended?
    xyz_aff = np.diag([2, 3, 4, 1])
    for size in (10,), (10, 2):
        data = np.random.normal(size=size)
        ni_img = nib.Nifti1Image(data, xyz_aff)
        # Default is aligned
        assert_raises(NiftiError, nifti2nipy, ni_img)
        hdr = ni_img.get_header()
        # The pixdim affine
        for label in 'scanner', 'aligned', 'talairach', 'mni':
            hdr.set_sform(xyz_aff, label)
            assert_raises(NiftiError, nifti2nipy, ni_img)
            hdr.set_sform(None)
            assert_raises(NiftiError, nifti2nipy, ni_img)
            hdr.set_sform(xyz_aff, label)
            assert_raises(NiftiError, nifti2nipy, ni_img)
            hdr.set_qform(None)


def test_load_cmaps():
    data = np.random.normal(size=range(7))
    xyz_aff = np.diag([2, 3, 4, 1])
    # Default with time-like
    ni_img = nib.Nifti1Image(data, xyz_aff)
    img = nifti2nipy(ni_img)
    exp_cmap = AT(CS('ijktuvw', name='voxels'),
                  aligned_csm(7),
                  np.diag([2, 3, 4, 1, 1, 1, 1, 1]))
    assert_equal(img.coordmap, exp_cmap)
    # xyzt_units sets time axis name
    hdr = ni_img.get_header()
    xyz_names = aligned_csm(3).coord_names
    full_aff = exp_cmap.affine
    reduced_data = data[:, :, :, 1:2, ...]
    for t_like, units, scaling in (
        ('t', 'sec', 1),
        ('t', 'msec', 1/1000.),
        ('t', 'usec', 1/1000000.),
        ('hz', 'hz', 1),
        ('ppm', 'ppm', 1),
        ('rads', 'rads', 1)):
        hdr.set_xyzt_units('mm', units)
        img = nifti2nipy(ni_img)
        in_cs = CS(('i', 'j', 'k', t_like, 'u', 'v', 'w'), name='voxels')
        out_cs = CS(xyz_names + (t_like, 'u', 'v', 'w'), name='aligned')
        if scaling == 1:
            exp_aff = full_aff
        else:
            diag = np.ones((8,))
            diag[3] = scaling
            exp_aff = np.dot(np.diag(diag), full_aff)
        exp_cmap = AT(in_cs, out_cs, exp_aff)
        assert_equal(img.coordmap, exp_cmap)
        assert_array_equal(img.get_data(), data)
        # Even if the image axis length is 1, we keep out time dimension, if
        # there is specific scaling implying time-like
        ni_img_t = nib.Nifti1Image(reduced_data, xyz_aff, hdr)
        img = nifti2nipy(ni_img_t)
        assert_equal(img.coordmap, exp_cmap)
        assert_array_equal(img.get_data(), reduced_data)


def test_load_no_time():
    # Without setting anything else, length 1 at position 3 makes time go away
    ns_dims = (5, 6, 7)
    xyz_aff = np.diag([2, 3, 4, 1])
    xyz_names = aligned_csm(3).coord_names[:3]
    in_names = tuple('ijkuvw')
    out_names = xyz_names + tuple('uvw')
    for n_ns in 1, 2, 3:
        ndim = 3 + n_ns
        data = np.random.normal(size=(2, 3, 4, 1) + ns_dims[:n_ns])
        ni_img_no_t = nib.Nifti1Image(data, xyz_aff)
        cmap_no_t = AT(CS(in_names[:ndim], name='voxels'),
                       CS(out_names[:ndim], name='aligned'),
                       np.diag([2, 3, 4] + [1] * n_ns + [1]))
        img = nifti2nipy(ni_img_no_t)
        assert_equal(img.coordmap, cmap_no_t)
    # We add do time if 4th axis of length 1 is the last axis
    data41 = np.zeros((3, 4, 5, 1))
    ni_img_41 = nib.Nifti1Image(data41, xyz_aff)
    cmap_41 = AT(CS('ijkt', name='voxels'),
                   CS(xyz_names + ('t',), name='aligned'),
                   np.diag([2, 3, 4, 1, 1]))
    img = nifti2nipy(ni_img_41)
    assert_equal(img.coordmap, cmap_41)


def test_load_toffset():
    # Test toffset gets set into affine only for time
    data = np.random.normal(size=range(5))
    xyz_aff = np.diag([2, 3, 4, 1])
    # Default with time-like and no toffset
    ni_img = nib.Nifti1Image(data, xyz_aff)
    hdr = ni_img.get_header()
    img = nifti2nipy(ni_img)
    exp_aff = np.diag([2., 3, 4, 1, 1, 1])
    in_cs = CS('ijktu', name='voxels')
    xyz_names = aligned_csm(3).coord_names
    out_cs = CS(xyz_names + tuple('tu'), name='aligned')
    assert_equal(hdr['toffset'], 0)
    assert_equal(img.coordmap, AT(in_cs, out_cs, exp_aff))
    # Set toffset and expect in affine
    hdr['toffset'] = 42
    exp_aff[3, -1] = 42
    assert_equal(nifti2nipy(ni_img).coordmap, AT(in_cs, out_cs, exp_aff))
    # Make time axis into hz and expect not to see toffset
    hdr.set_xyzt_units('mm', 'hz')
    in_cs_hz = CS(('i', 'j', 'k', 'hz', 'u'), name='voxels')
    out_cs_hz = CS(xyz_names + ('hz', 'u'), name='aligned')
    exp_aff[3, -1] = 0
    assert_equal(nifti2nipy(ni_img).coordmap, AT(in_cs_hz, out_cs_hz, exp_aff))


def test_load_spaces():
    # Test spaces get read correctly
    shape = np.array((6, 5, 4, 3, 2))
    zooms = np.array((2, 3, 4, 5, 6))
    data = np.random.normal(size=shape)
    # Default with no affine in header, or in image
    ni_img = nib.Nifti1Image(data, None)
    hdr = ni_img.get_header()
    hdr.set_zooms(zooms)
    # Expected affine is from the pixdims and the center of the image.  Default
    # is also flipped X.
    offsets = (1 - shape[:3]) / 2. * zooms[:3] * (-1, 1, 1)
    exp_aff = from_matvec(np.diag([-2, 3, 4, 5, 6]),
                          list(offsets) + [0, 0])
    in_cs = CS('ijktu', name='voxels')
    exp_cmap = AT(in_cs, unknown_csm(5), exp_aff)
    assert_equal(nifti2nipy(ni_img).coordmap, exp_cmap)
    an_aff = from_matvec(np.diag([1.1, 2.2, 3.3]), [10, 11, 12])
    exp_aff = from_matvec(np.diag([1.1, 2.2, 3.3, 5, 6]), [10, 11, 12, 0, 0])
    for label, csm in (('scanner', scanner_csm),
                       ('aligned', aligned_csm),
                       ('talairach', talairach_csm),
                       ('mni', mni_csm)):
        hdr.set_sform(an_aff, label)
        assert_equal(nifti2nipy(ni_img).coordmap, AT(in_cs, csm(5), exp_aff))


def test_mm_scaling():
    # Test the micron and meter scale the affine right
    data = np.random.normal(size=range(4))
    xyz_aff = from_matvec(np.diag([2, 3, 4]), [11, 12, 13])
    exp_aff = from_matvec(np.diag([2, 3, 4, 1]), [11, 12, 13, 0])
    in_cs = CS('ijkt', name='voxels')
    out_cs = aligned_csm(4)
    # No space scaling
    ni_img = nib.Nifti1Image(data, xyz_aff)
    hdr = ni_img.get_header()
    assert_equal(hdr.get_xyzt_units(), ('unknown', 'unknown'))
    assert_equal(nifti2nipy(ni_img).coordmap, AT(in_cs, out_cs, exp_aff))
    # mm is assumed
    hdr.set_xyzt_units('mm')
    assert_equal(nifti2nipy(ni_img).coordmap, AT(in_cs, out_cs, exp_aff))
    # microns !
    hdr.set_xyzt_units('micron')
    scaler = np.diag([1 / 1000., 1 / 1000., 1 / 1000., 1, 1])
    assert_equal(nifti2nipy(ni_img).coordmap,
                 AT(in_cs, out_cs, np.dot(scaler, exp_aff)))
    # mm again !  This test implicitly asserts that the nifti image affine is
    # not being changed by the conversion routine, otherwise we'd pick up the
    # microns scaling above.
    hdr.set_xyzt_units('mm')
    assert_equal(nifti2nipy(ni_img).coordmap, AT(in_cs, out_cs, exp_aff))
    # meters !
    hdr.set_xyzt_units('meter')
    scaler = np.diag([1000., 1000., 1000., 1, 1])
    assert_equal(nifti2nipy(ni_img).coordmap,
                 AT(in_cs, out_cs, np.dot(scaler, exp_aff)))


def test_load_dim_info():
    # Test freq, phase, slice get set correctly on load
    data = np.random.normal(size=range(3))
    xyz_aff = from_matvec(np.diag([2, 3, 4]), [11, 12, 13])
    in_cs = CS('ijk', name='voxels')
    out_cs = aligned_csm(3)
    # Just confirm that the default leads to no axis renaming
    ni_img = nib.Nifti1Image(data, xyz_aff)
    hdr = ni_img.get_header()
    assert_equal(hdr.get_dim_info(), (None, None, None))
    assert_equal(nifti2nipy(ni_img).coordmap, AT(in_cs, out_cs, xyz_aff))
    # But now...
    hdr.set_dim_info(freq=1)
    assert_equal(nifti2nipy(ni_img).coordmap,
                 AT(CS(('i', 'freq', 'k'), "voxels"), out_cs, xyz_aff))
    hdr.set_dim_info(freq=2)
    assert_equal(nifti2nipy(ni_img).coordmap,
                 AT(CS(('i', 'j', 'freq'), "voxels"), out_cs, xyz_aff))
    hdr.set_dim_info(phase=1)
    assert_equal(nifti2nipy(ni_img).coordmap,
                 AT(CS(('i', 'phase', 'k'), "voxels"), out_cs, xyz_aff))
    hdr.set_dim_info(slice=0)
    assert_equal(nifti2nipy(ni_img).coordmap,
                 AT(CS(('slice', 'j', 'k'), "voxels"), out_cs, xyz_aff))
    hdr.set_dim_info(freq=1, phase=0, slice=2)
    assert_equal(nifti2nipy(ni_img).coordmap,
                 AT(CS(('phase', 'freq', 'slice'), "voxels"), out_cs, xyz_aff))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Example analyzing the FIAC dataset with NIPY.

* Single run models with per-voxel AR(1).
* Cross-run, within-subject models with optimal effect estimates.
* Cross-subject models using fixed / random effects variance ratios.
* Permutation testing for inference on cross-subject result.

See ``parallel_run.py`` for a rig to run these analysis in parallel using the
IPython parallel machinery.

This script needs the pre-processed FIAC data.  See ``README.txt`` and
``fiac_util.py`` for details.

See ``examples/labs/need_data/first_level_fiac.py`` for an alternative approach
to some of these analyses.
"""
#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from __future__ import print_function # Python 2/3 compatibility

# Stdlib
from tempfile import NamedTemporaryFile
from os.path import join as pjoin
from copy import copy
import warnings

# Third party
import numpy as np

# From NIPY
from nipy.algorithms.statistics.api import (OLSModel, ARModel, make_recarray,
                                            isestimable)
from nipy.modalities.fmri.fmristat import hrf as delay
from nipy.modalities.fmri import design, hrf
from nipy.io.api import load_image, save_image
from nipy.core import api
from nipy.core.api import Image
from nipy.core.image.image import rollimg

from nipy.algorithms.statistics import onesample

# Local
import fiac_util as futil
reload(futil)  # while developing interactively

#-----------------------------------------------------------------------------
# Globals
#-----------------------------------------------------------------------------

SUBJECTS = tuple(range(5) + range(6, 16)) # No data for subject 5
RUNS = tuple(range(1, 5))
DESIGNS = ('event', 'block')
CONTRASTS = ('speaker_0', 'speaker_1',
             'sentence_0', 'sentence_1',
             'sentence:speaker_0',
             'sentence:speaker_1')

GROUP_MASK = futil.load_image_fiac('group', 'mask.nii')
TINY_MASK = np.zeros(GROUP_MASK.shape, np.bool)
TINY_MASK[30:32,40:42,30:32] = 1

#-----------------------------------------------------------------------------
# Public functions
#-----------------------------------------------------------------------------

# For group analysis

def run_model(subj, run):
    """
    Single subject fitting of FIAC model
    """
    #----------------------------------------------------------------------
    # Set initial parameters of the FIAC dataset
    #----------------------------------------------------------------------
    # Number of volumes in the fMRI data
    nvol = 191
    # The TR of the experiment
    TR = 2.5
    # The time of the first volume
    Tstart = 0.0
    # The array of times corresponding to each volume in the fMRI data
    volume_times = np.arange(nvol) * TR + Tstart
    # This recarray of times has one column named 't'.  It is used in the
    # function design.event_design to create the design matrices.
    volume_times_rec = make_recarray(volume_times, 't')
    # Get a path description dictionary that contains all the path data relevant
    # to this subject/run
    path_info = futil.path_info_run(subj,run)

    #----------------------------------------------------------------------
    # Experimental design
    #----------------------------------------------------------------------

    # Load the experimental description from disk.  We have utilities in futil
    # that reformat the original FIAC-supplied format into something where the
    # factorial structure of the design is more explicit.  This has already
    # been run once, and get_experiment_initial() will simply load the
    # newly-formatted design description files (.csv) into record arrays.
    experiment, initial = futil.get_experiment_initial(path_info)

    # Create design matrices for the "initial" and "experiment" factors, saving
    # the default contrasts.

    # The function event_design will create design matrices, which in the case
    # of "experiment" will have num_columns = (# levels of speaker) * (# levels
    # of sentence) * len(delay.spectral) = 2 * 2 * 2 = 8. For "initial", there
    # will be (# levels of initial) * len([hrf.glover]) = 1 * 1 = 1.

    # Here, delay.spectral is a sequence of 2 symbolic HRFs that are described
    # in:
    #
    # Liao, C.H., Worsley, K.J., Poline, J-B., Aston, J.A.D., Duncan, G.H.,
    #    Evans, A.C. (2002). \'Estimating the delay of the response in fMRI
    #    data.\' NeuroImage, 16:593-606.

    # The contrast definitions in ``cons_exper`` are a dictionary with keys
    # ['constant_0', 'constant_1', 'speaker_0', 'speaker_1', 'sentence_0',
    # 'sentence_1', 'sentence:speaker_0', 'sentence:speaker_1'] representing the
    # four default contrasts: constant, main effects + interactions, each
    # convolved with 2 HRFs in delay.spectral. For example, sentence:speaker_0
    # is the interaction of sentence and speaker convolved with the first (=0)
    # of the two HRF basis functions, and sentence:speaker_1 is the interaction
    # convolved with the second (=1) of the basis functions.

    # XXX use the hrf __repr__ for naming contrasts
    X_exper, cons_exper = design.event_design(experiment, volume_times_rec,
                                              hrfs=delay.spectral)

    # The contrasts for 'initial' are ignored as they are "uninteresting" and
    # are included in the model as confounds.
    X_initial, _ = design.event_design(initial, volume_times_rec,
                                       hrfs=[hrf.glover])

    # In addition to factors, there is typically a "drift" term. In this case,
    # the drift is a natural cubic spline with a not at the midpoint
    # (volume_times.mean())
    vt = volume_times # shorthand
    drift = np.array( [vt**i for i in range(4)] +
                      [(vt-vt.mean())**3 * (np.greater(vt, vt.mean()))] )
    for i in range(drift.shape[0]):
        drift[i] /= drift[i].max()

    # We transpose the drift so that its shape is (nvol,5) so that it will have
    # the same number of rows as X_initial and X_exper.
    drift = drift.T

    # There are helper functions to create these drifts: design.fourier_basis,
    # design.natural_spline.  Therefore, the above is equivalent (except for
    # the normalization by max for numerical stability) to
    #
    # >>> drift = design.natural_spline(t, [volume_times.mean()])

    # Stack all the designs, keeping the new contrasts which has the same keys
    # as cons_exper, but its values are arrays with 15 columns, with the
    # non-zero entries matching the columns of X corresponding to X_exper
    X, cons = design.stack_designs((X_exper, cons_exper),
                                   (X_initial, {}),
                                   (drift, {}))

    # Sanity check: delete any non-estimable contrasts
    for k in cons.keys():
        if not isestimable(cons[k], X):
            del(cons[k])
            warnings.warn("contrast %s not estimable for this run" % k)

    # The default contrasts are all t-statistics.  We may want to output
    # F-statistics for 'speaker', 'sentence', 'speaker:sentence' based on the
    # two coefficients, one for each HRF in delay.spectral

    cons['speaker'] = np.vstack([cons['speaker_0'], cons['speaker_1']])
    cons['sentence'] = np.vstack([cons['sentence_0'], cons['sentence_1']])
    cons['sentence:speaker'] = np.vstack([cons['sentence:speaker_0'],
                                          cons['sentence:speaker_1']])

    #----------------------------------------------------------------------
    # Data loading
    #----------------------------------------------------------------------

    # Load in the fMRI data, saving it as an array.  It is transposed to have
    # time as the first dimension, i.e. fmri[t] gives the t-th volume.
    fmri_im = futil.get_fmri(path_info) # an Image
    fmri_im = rollimg(fmri_im, 't')
    fmri = fmri_im.get_data() # now, it's an ndarray

    nvol, volshape = fmri.shape[0], fmri.shape[1:]
    nx, sliceshape = volshape[0], volshape[1:]

    #----------------------------------------------------------------------
    # Model fit
    #----------------------------------------------------------------------

    # The model is a two-stage model, the first stage being an OLS (ordinary
    # least squares) fit, whose residuals are used to estimate an AR(1)
    # parameter for each voxel.
    m = OLSModel(X)
    ar1 = np.zeros(volshape)

    # Fit the model, storing an estimate of an AR(1) parameter at each voxel
    for s in range(nx):
        d = np.array(fmri[:,s])
        flatd = d.reshape((d.shape[0], -1))
        result = m.fit(flatd)
        ar1[s] = ((result.resid[1:] * result.resid[:-1]).sum(0) /
                  (result.resid**2).sum(0)).reshape(sliceshape)

    # We round ar1 to nearest one-hundredth and group voxels by their rounded
    # ar1 value, fitting an AR(1) model to each batch of voxels.

    # XXX smooth here?
    # ar1 = smooth(ar1, 8.0)
    ar1 *= 100
    ar1 = ar1.astype(np.int) / 100.

    # We split the contrasts into F-tests and t-tests.
    # XXX helper function should do this
    fcons = {}; tcons = {}
    for n, v in cons.items():
        v = np.squeeze(v)
        if v.ndim == 1:
            tcons[n] = v
        else:
            fcons[n] = v

    # Setup a dictionary to hold all the output
    # XXX ideally these would be memmap'ed Image instances
    output = {}
    for n in tcons:
        tempdict = {}
        for v in ['sd', 't', 'effect']:
            tempdict[v] = np.memmap(NamedTemporaryFile(prefix='%s%s.nii'
                                    % (n,v)), dtype=np.float,
                                    shape=volshape, mode='w+')
        output[n] = tempdict

    for n in fcons:
        output[n] = np.memmap(NamedTemporaryFile(prefix='%s%s.nii'
                                    % (n,v)), dtype=np.float,
                                    shape=volshape, mode='w+')

    # Loop over the unique values of ar1
    for val in np.unique(ar1):
        armask = np.equal(ar1, val)
        m = ARModel(X, val)
        d = fmri[:,armask]
        results = m.fit(d)

        # Output the results for each contrast
        for n in tcons:
            resT = results.Tcontrast(tcons[n])
            output[n]['sd'][armask] = resT.sd
            output[n]['t'][armask] = resT.t
            output[n]['effect'][armask] = resT.effect
        for n in fcons:
            output[n][armask] = results.Fcontrast(fcons[n]).F

    # Dump output to disk
    odir = futil.output_dir(path_info,tcons,fcons)
    # The coordmap for a single volume in the time series
    vol0_map = fmri_im[0].coordmap
    for n in tcons:
        for v in ['t', 'sd', 'effect']:
            im = Image(output[n][v], vol0_map)
            save_image(im, pjoin(odir, n, '%s.nii' % v))
    for n in fcons:
        im = Image(output[n], vol0_map)
        save_image(im, pjoin(odir, n, "F.nii"))


def fixed_effects(subj, design):
    """ Fixed effects (within subject) for FIAC model

    Finds run by run estimated model results, creates fixed effects results
    image per subject.

    Parameters
    ----------
    subj : int
        subject number 0..15 inclusive
    design : {'block', 'event'}
        design type
    """
    # First, find all the effect and standard deviation images
    # for the subject and this design type
    path_dict = futil.path_info_design(subj, design)
    rootdir = path_dict['rootdir']
    # The output directory
    fixdir = pjoin(rootdir, "fixed")
    # Fetch results images from run estimations
    results = futil.results_table(path_dict)
    # Get our hands on the relevant coordmap to save our results
    coordmap = futil.load_image_fiac("fiac_%02d" % subj,
                                     "wanatomical.nii").coordmap
    # Compute the "fixed" effects for each type of contrast
    for con in results:
        fixed_effect = 0
        fixed_var = 0
        for effect, sd in results[con]:
            effect = load_image(effect).get_data()
            sd = load_image(sd).get_data()
            var = sd ** 2

            # The optimal, in terms of minimum variance, combination of the
            # effects has weights 1 / var
            #
            # XXX regions with 0 variance are set to 0
            # XXX do we want this or np.nan?
            ivar = np.nan_to_num(1. / var)
            fixed_effect += effect * ivar
            fixed_var += ivar

        # Now, compute the fixed effects variance and t statistic
        fixed_sd = np.sqrt(fixed_var)
        isd = np.nan_to_num(1. / fixed_sd)
        fixed_t = fixed_effect * isd

        # Save the results
        odir = futil.ensure_dir(fixdir, con)
        for a, n in zip([fixed_effect, fixed_sd, fixed_t],
                        ['effect', 'sd', 't']):
            im = api.Image(a, copy(coordmap))
            save_image(im, pjoin(odir, '%s.nii' % n))


def group_analysis(design, contrast):
    """ Compute group analysis effect, t, sd for `design` and `contrast`

    Saves to disk in 'group' analysis directory

    Parameters
    ----------
    design : {'block', 'event'}
    contrast : str
        contrast name
    """
    array = np.array # shorthand
    # Directory where output will be written
    odir = futil.ensure_dir(futil.DATADIR, 'group', design, contrast)

    # Which subjects have this (contrast, design) pair?
    subj_con_dirs = futil.subj_des_con_dirs(design, contrast)
    if len(subj_con_dirs) == 0:
        raise ValueError('No subjects for %s, %s' % (design, contrast))

    # Assemble effects and sds into 4D arrays
    sds = []
    Ys = []
    for s in subj_con_dirs:
        sd_img = load_image(pjoin(s, "sd.nii"))
        effect_img = load_image(pjoin(s, "effect.nii"))
        sds.append(sd_img.get_data())
        Ys.append(effect_img.get_data())
    sd = array(sds)
    Y = array(Ys)

    # This function estimates the ratio of the fixed effects variance
    # (sum(1/sd**2, 0)) to the estimated random effects variance
    # (sum(1/(sd+rvar)**2, 0)) where rvar is the random effects variance.

    # The EM algorithm used is described in:
    #
    # Worsley, K.J., Liao, C., Aston, J., Petre, V., Duncan, G.H.,
    #    Morales, F., Evans, A.C. (2002). \'A general statistical
    #    analysis for fMRI data\'. NeuroImage, 15:1-15
    varest = onesample.estimate_varatio(Y, sd)
    random_var = varest['random']

    # XXX - if we have a smoother, use
    # random_var = varest['fixed'] * smooth(varest['ratio'])

    # Having estimated the random effects variance (and possibly smoothed it),
    # the corresponding estimate of the effect and its variance is computed and
    # saved.

    # This is the coordmap we will use
    coordmap = futil.load_image_fiac("fiac_00","wanatomical.nii").coordmap

    adjusted_var = sd**2 + random_var
    adjusted_sd = np.sqrt(adjusted_var)

    results = onesample.estimate_mean(Y, adjusted_sd) 
    for n in ['effect', 'sd', 't']:
        im = api.Image(results[n], copy(coordmap))
        save_image(im, pjoin(odir, "%s.nii" % n))


def group_analysis_signs(design, contrast, mask, signs=None):
    """ Refit the EM model with a vector of signs.

    Used in the permutation tests.

    Returns the maximum of the T-statistic within mask

    Parameters
    ----------
    design: one of 'block', 'event'
    contrast: str
        name of contrast to estimate
    mask : ``Image`` instance or array-like
        image containing mask, or array-like
    signs: ndarray, optional
         Defaults to np.ones. Should have shape (*,nsubj)
         where nsubj is the number of effects combined in the group analysis.

    Returns
    -------
    minT: np.ndarray, minima of T statistic within mask, one for each
         vector of signs
    maxT: np.ndarray, maxima of T statistic within mask, one for each
         vector of signs
    """
    if api.is_image(mask):
        maska = mask.get_data()
    else:
        maska = np.asarray(mask)
    maska = maska.astype(np.bool)

    # Which subjects have this (contrast, design) pair?
    subj_con_dirs = futil.subj_des_con_dirs(design, contrast)

    # Assemble effects and sds into 4D arrays
    sds = []
    Ys = []
    for s in subj_con_dirs:
        sd_img = load_image(pjoin(s, "sd.nii"))
        effect_img = load_image(pjoin(s, "effect.nii"))
        sds.append(sd_img.get_data()[maska])
        Ys.append(effect_img.get_data()[maska])
    sd = np.array(sds)
    Y = np.array(Ys)

    if signs is None:
        signs = np.ones((1, Y.shape[0]))

    maxT = np.empty(signs.shape[0])
    minT = np.empty(signs.shape[0])

    for i, sign in enumerate(signs):
        signY = sign[:,np.newaxis] * Y
        varest = onesample.estimate_varatio(signY, sd)
        random_var = varest['random']

        adjusted_var = sd**2 + random_var
        adjusted_sd = np.sqrt(adjusted_var)

        results = onesample.estimate_mean(Y, adjusted_sd) 
        T = results['t']
        minT[i], maxT[i] = np.nanmin(T), np.nanmax(T)
    return minT, maxT


def permutation_test(design, contrast, mask=GROUP_MASK, nsample=1000):
    """
    Perform a permutation (sign) test for a given design type and
    contrast. It is a Monte Carlo test because we only sample nsample
    possible sign arrays.

    Parameters
    ----------
    design: str
        one of ['block', 'event']
    contrast : str
        name of contrast to estimate
    mask : ``Image`` instance or array-like, optional
        image containing mask, or array-like
    nsample: int, optional
        number of permutations

    Returns
    -------
    min_vals: np.ndarray
    max_vals: np.ndarray
    """
    subj_con_dirs = futil.subj_des_con_dirs(design, contrast)
    nsubj = len(subj_con_dirs)
    if nsubj == 0:
        raise ValueError('No subjects have %s, %s' % (design, contrast))
    signs = 2*np.greater(np.random.sample(size=(nsample, nsubj)), 0.5) - 1
    min_vals, max_vals = group_analysis_signs(design, contrast, mask, signs)
    return min_vals, max_vals


def run_run_models(subject_nos=SUBJECTS, run_nos = RUNS):
    """ Simple serial run of all the within-run models """
    for subj in subject_nos:
        for run in run_nos:
            try:
                run_model(subj, run)
            except IOError:
                print('Skipping subject %d, run %d' % (subj, run))


def run_fixed_models(subject_nos=SUBJECTS, designs=DESIGNS):
    """ Simple serial run of all the within-subject models """
    for subj in subject_nos:
        for design in designs:
            try:
                fixed_effects(subj, design)
            except IOError:
                print('Skipping subject %d, design %s' % (subj, design))


def run_group_models(designs=DESIGNS, contrasts=CONTRASTS):
    """ Simple serial run of all the across-subject models """
    for design in designs:
        for contrast in contrasts:
            group_analysis(design, contrast)


if __name__ == '__main__':
    pass
    # Sanity check while debugging
    #permutation_test('block','sentence_0',mask=TINY_MASK,nsample=3)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This figure is meant to represent an event-type design with
events at random times centered [0,2,4,6,8] and random
amplitudes centered [0,1.1,2.3,0.9,0.3].
"""

import pylab, matplotlib
import numpy as np


for t, y in zip([0,4,8,12,16], [0,1.1,2.3,0.9,0.3]):
    dt = np.array([-0.5,0.5,0.5,-0.5])
    dy = np.array([-0.1,-0.1,0.1,0.1])
    pylab.fill(t+dt,y+dy, 'r')

a = pylab.gca()
a.set_yticks([0,2])
a.set_xlim([-1,18])
a.set_xlabel('Time')
a.set_ylabel('Amplitude')




#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
This example illustrates  the impact of using a mixed-effects model
for the detection of the effects, when the first-level variance is known:
If the first level variance is very variable across observations, then taking
it into account gives more relibale detections, as seen in an ROC curve.

Requires matplotlib.

Author: Bertrand Thirion, 2012
"""
print(__doc__)

import numpy as np
import matplotlib.pyplot as plt

from nipy.algorithms.statistics.mixed_effects_stat import (
    generate_data, one_sample_ttest, t_stat)

# generate the data
N, P = 15, 500
V1 = np.random.randn(N, P) ** 2
effects = 0.5 * (np.random.randn(P) > 0)
Y = generate_data(np.ones(N), effects, .25, V1)

# compute the statistics
T1 = one_sample_ttest(Y, V1, n_iter=5)
T1 = [T1[effects == x] for x  in np.unique(effects)]
T2 = [t_stat(Y)[effects == x] for x  in np.unique(effects)]

# Derive ROC curves
ROC1 = np.array([np.sum(T1[1] > - x) for x  in np.sort(- T1[0])])\
    * 1. / T1[1].size
ROC2 = np.array([np.sum(T2[1] > - x) for x  in np.sort(- T2[0])])\
    * 1. / T1[1].size

# make a figure
FIG = plt.figure(figsize=(10, 5))
AX = FIG.add_subplot(121)
AX.plot(np.linspace(0, 1, len(ROC1)), ROC1, label='mixed effects')
AX.plot(np.linspace(0, 1, len(ROC2)), ROC2, label='t test')
AX.set_xlabel('false positives')
AX.set_ylabel('true positives')
AX.set_title('ROC curves for the detection of effects', fontsize=12)
AX.legend(loc='lower right')
AX = FIG.add_subplot(122)
AX.boxplot(T1, positions=[-0.1, .9])
AX.boxplot(T2, positions=[0.1, 1.1])
AX.set_xticks([0, 1])
AX.set_xlabel('simulated effects')
AX.set_ylabel('decision statistic')
AX.set_title('left: mixed effects model, \n right: standard t test',
             fontsize=12)
plt.show()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# models - Statistical Models
#

__docformat__ = 'restructuredtext'

from .info import __doc__

from . import model
from . import regression
from . import glm

__all__ = filter(lambda s:not s.startswith('_'),dir())

from nipy.testing import Tester
test = Tester().test



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This module contains the specification of 'hierarchical ROI' object,
Which is used in spatial models of the library such as structural analysis

The connection with other classes is not completely satisfactory at the moment:
there should be some intermediate classes between 'Fields' and 'hroi'

Author : Bertrand Thirion, 2009-2011
         Virgile Fritsch <virgile.fritsch@inria.fr>

"""

import numpy as np

from nipy.algorithms.graph.graph import WeightedGraph
from nipy.algorithms.graph.forest import Forest
from nipy.algorithms.graph.field import field_from_coo_matrix_and_data
from .mroi import SubDomains

NINF = - np.inf


def hroi_agglomeration(input_hroi, criterion='size', smin=0):
    """Performs an agglomeration then a selection of regions
    so that a certain size or volume criterion is satisfied.

    Parameters
    ----------
    input_hroi: HierarchicalROI instance
      The input hROI
    criterion: str, optional
      To be chosen among 'size' or 'volume'
    smin: float, optional
      The applied criterion

    Returns
    -------
    output_hroi:  HierarchicalROI instance
    """
    if criterion not in ['size', 'volume']:
        return ValueError('unknown criterion')
    output_hroi = input_hroi.copy()
    k = 2 * output_hroi.k
    if criterion == 'size':
        value = output_hroi.get_size()
    if criterion == 'volume':
        value = output_hroi.get_volume()

    # iteratively agglomerate regions that are too small
    while k > output_hroi.k:
        k = output_hroi.k
        # regions agglomeration
        output_hroi.merge_ascending(output_hroi.get_id()[value <= smin])
        # suppress parents nodes having only one child
        output_hroi.merge_descending()
        # early stopping 1
        if output_hroi.k == 0:
            break
        # early stopping 2
        if criterion == 'size':
            value = output_hroi.get_size()
        if criterion == 'volume':
            value = output_hroi.get_volume()
        if value.max() < smin:
            break

    # finally remove those regions for which the criterion cannot be matched
    output_hroi.select_roi(output_hroi.get_id()[value > smin])
    return output_hroi


def HROI_as_discrete_domain_blobs(domain, data, threshold=NINF, smin=0,
                                  criterion='size'):
    """Instantiate an HierarchicalROI as the blob decomposition
    of data in a certain domain.

    Parameters
    ----------
    domain : discrete_domain.StructuredDomain instance,
      Definition of the spatial context.
    data : array of shape (domain.size)
      The corresponding data field.
    threshold : float, optional
      Thresholding level.
    criterion : string, optional
      To be chosen among 'size' or 'volume'.
    smin: float, optional
      A threshold on the criterion.

    Returns
    -------
    nroi: HierachicalROI instance with a `signal` feature.

    """
    if threshold > data.max():
        # return an empty HROI structure
        label = - np.ones(data.shape)
        parents = np.array([])
        return HierarchicalROI(domain, label, parents)

    # check size
    df = field_from_coo_matrix_and_data(domain.topology, data)
    idx, parents, label = df.threshold_bifurcations(th=threshold)
    nroi = HierarchicalROI(domain, label, parents)
    # create a signal feature
    data = np.ravel(data)
    signal = [data[nroi.select_id(id, roi=False)] for id in nroi.get_id()]
    nroi.set_feature('signal', signal)
    # agglomerate regions in order to compact the structure if necessary
    nroi = hroi_agglomeration(nroi, criterion=criterion, smin=smin)
    return nroi


def HROI_from_watershed(domain, data, threshold=NINF):
    """Instantiate an HierarchicalROI as the watershed of a certain dataset

    Parameters
    ----------
    domain: discrete_domain.StructuredDomain instance
      Definition of the spatial context.
    data: array of shape (domain.size)
      The corresponding data field.
    threshold: float, optional
      Thresholding level.

    Returns
    -------
    nroi : ``HierarchichalROI`` instance
        The HierachicalROI instance with a ``seed`` feature.
    """
    if threshold > data.max():
        # return an empty HROI structure
        label = - np.ones(data.shape)
        parents = np.array([])
        return HierarchicalROI(domain, label, parents)

    df = field_from_coo_matrix_and_data(domain.topology, data)
    idx, label = df.custom_watershed(0, threshold)
    parents = np.arange(idx.size).astype(int)
    nroi = HierarchicalROI(domain, label, parents)

    nroi.set_roi_feature('seed', idx)
    return nroi


########################################################################
# Hierarchical ROI
########################################################################
class HierarchicalROI(SubDomains):
    """Class that handles hierarchical ROIs

    Parameters
    ----------
    k : int
      Number of ROI in the SubDomains object
    label : array of shape (domain.size), dtype=np.int
      An array use to define which voxel belongs to which ROI.
      The label values greater than -1 correspond to subregions
      labelling. The labels are recomputed so as to be consecutive
      integers.
      The labels should not be accessed outside this class. One has to
      use the API mapping methods instead.
    features : dict {str: list of object, length=self.k}
      Describe the voxels features, grouped by ROI
    roi_features : dict {str: array-like, shape=(self.k, roi_feature_dim)
      Describe the ROI features. A special feature, `id`, is read-only and
      is used to give an unique identifier for region, which is persistent
      through the MROI objects manipulations. On should access the different
      ROI's features using ids.
    parents : np.ndarray, shape(self.k)
      self.parents[i] is the index of the parent of the i-th ROI.

    TODO: have the parents as a list of id rather than a list of indices.
    """

    def __init__(self, domain, label, parents, id=None):
        """Building the HierarchicalROI
        """
        SubDomains.__init__(self, domain, label, id=id)
        self.parents = np.ravel(parents).astype(np.int)

    ###
    # Getters for very basic features or roi features
    ###
    def get_volume(self, id=None, ignore_children=True):
        """Get ROI volume

        Parameters
        ----------
        id: any hashable type, optional
          Id of the ROI from which we want to get the volume.
          Can be None (default) if we want all ROIs's volumes.
        ignore_children : bool, optional
          Specify if the volume of the node should include
          (ignore_children = False) or not the one of its children
          (ignore_children = True).

        Returns
        -------
        volume : float
          if an id is provided,
             or list of float
          if no id provided (default)
        """
        if ignore_children:
            # volume of the children is not included
            volume = SubDomains.get_volume(self, id)
        else:
            # volume of the children is included
            if id is not None:
                volume = SubDomains.get_volume(self, id)
                desc = self.make_forest().get_descendants(
                    self.select_id(id), exclude_self=True)
                # get children volume
                for k in desc:
                    volume = volume + SubDomains.get_volume(
                        self, self.get_id()[k])
            else:
                volume = []
                for id in self.get_id():
                    roi_volume = SubDomains.get_volume(self, id)
                    desc = self.make_forest().get_descendants(
                        self.select_id(id), exclude_self=True)
                    # get children volume
                    for k in desc:
                        roi_volume = roi_volume + SubDomains.get_volume(
                            self, self.get_id()[k])
                    volume.append(roi_volume)
        return volume

    def get_size(self, id=None, ignore_children=True):
        """Get ROI size (counted in terms of voxels)

        Parameters
        ----------
        id: any hashable type, optional
          Id of the ROI from which we want to get the size.
          Can be None (default) if we want all ROIs's sizes.
        ignore_children: bool, optional
          Specify if the size of the node should include
          (ignore_children = False) or not the one of its children
          (ignore_children = True).

        Returns
        -------
        size: int
          if an id is provided,
             or list of int
          if no id provided (default)

        """
        if ignore_children:
            # size of the children is not included
            size = SubDomains.get_size(self, id)
        else:
            # size of the children is included
            if id is not None:
                size = SubDomains.get_size(self, id)
                desc = self.make_forest().get_descendants(
                    self.select_id(id), exclude_self=True)
                # get children size
                for k in desc:
                    size = size + SubDomains.get_size(self, self.get_id()[k])
            else:
                size = []
                for id in self.get_id():
                    roi_size = SubDomains.get_size(self, id)
                    desc = self.make_forest().get_descendants(
                        self.select_id(id), exclude_self=True)
                    # get children size
                    for k in desc:
                        roi_size = roi_size + SubDomains.get_size(
                            self, self.get_id()[k])
                    size.append(roi_size)
        return size

    def select_roi(self, id_list):
        """Returns an instance of HROI with only the subset of chosen ROIs.

        The hierarchy is set accordingly.

        Parameters
        ----------
        id_list: list of id (any hashable type)
          The id of the ROI to be kept in the structure.

        """
        valid = np.asarray([int(i in id_list) for i in self.get_id()])
        if np.size(id_list) == 0:
            # handle the case of an empty selection
            new_parents = np.array([])
            self = HierarchicalROI(
                self.domain, -np.ones(self.label.size), np.array([]))
        else:
            # get new parents
            new_parents = Forest(self.k, self.parents).subforest(
                valid.astype(np.bool)).parents.astype(np.int)
        SubDomains.select_roi(self, id_list)
        self.parents = new_parents
        self.recompute_labels()

    def make_graph(self):
        """Output an nipy graph structure to represent the ROI hierarchy.

        """
        if self.k == 0:
            return None
        weights = np.ones(self.k)
        edges = (np.vstack((np.arange(self.k), self.parents))).T
        return WeightedGraph(self.k, edges, weights)

    def make_forest(self):
        """Output an nipy forest structure to represent the ROI hierarchy.

        """
        if self.k == 0:
            return None
        G = Forest(self.k, self.parents)
        return G

    def merge_ascending(self, id_list, pull_features=None):
        """Remove the non-valid ROIs by including them in
        their parents when it exists.

        Parameters
        ----------
        id_list: list of id (any hashable type)
          The id of the ROI to be merged into their parents.
          Nodes that are their own parent are unmodified.
        pull_features: list of str
          List of the ROI features that will be pooled from the children
          when they are merged into their parents. Otherwise, the receiving
          parent would keep its own ROI feature.
        """
        if pull_features is None:
            pull_features = []
        if self.k == 0:
            return
        id_list = [k for k in self.get_id() if k in id_list]

        # relabel maps old labels to new labels 
        relabel = np.arange(self.k)

        # merge nodes, one at a time
        for c_id in id_list:
            # define alias for clearer indexing
            c_pos = self.select_id(c_id)
            p_pos = self.parents[c_pos]
            p_id = self.get_id()[p_pos]
            
            if p_pos != c_pos:
                # this will be used in many places
                mask_pos = np.ones(self.k, np.bool)
                mask_pos[c_pos] = False
                
                # set new parents
                self.parents = self.parents[mask_pos]
                self.parents[self.parents == c_pos] = p_pos
                self.parents[self.parents > c_pos] -=  1
                self.k -= 1

                # merge labels
                relabel[relabel == c_id] = p_id

                # compute new features
                for fid in self.features.keys():
                    # replace feature
                    # (without the API since self is in an inconsistent state)
                    dj = self.get_feature(fid)
                    dj[p_pos] = np.hstack((dj[self.select_id(c_id)], 
                                           dj[self.select_id(p_id)]))
                    del dj[c_pos]
                    self.features[fid] = dj

                # compute new roi features
                for fid in self.roi_features.keys():
                    dj = self.get_roi_feature(fid)
                    if fid in pull_features:
                        # modify only if `pull` requested
                        dj[p_pos] = dj[c_pos]
                    self.roi_features[fid] = dj[mask_pos]

        # update the labels        
        self.label[self.label > -1] = relabel[self.label[self.label > - 1]]
        self.recompute_labels()

    def merge_descending(self, pull_features=None):
        """ Remove the items with only one son by including them in their son

        Parameters
        ----------
        methods indicates the way possible features are dealt with
        (not implemented yet)

        Caveat
        ------
        if roi_features have been defined, they will be removed
        """
        if pull_features is None:
            pull_features = []

        if self.k == 0:
            return
        
        # relabel maps old labels to new labels 
        relabel = np.arange(self.k)

        # merge nodes, one at a time
        id_list = self.get_id()[:: - 1]
        
        for p_id in id_list:
            p_pos = self.select_id(p_id)
            p_children = np.nonzero(self.parents == p_pos)[0]
            
            if p_pos in p_children:
                # remove current node from its children list
                p_children = p_children[p_children != p_pos]

            if p_children.size == 1:
                # merge node if it has only one child
                c_pos = p_children[0]
                c_id = self.get_id()[c_pos]
                mask_pos = np.ones(self.k, np.bool)
                mask_pos[p_pos] = False

                # set new parents
                self.parents[c_pos] = self.parents[p_pos]
                if self.parents[c_pos] == p_pos:
                    self.parents[c_pos] = c_pos
                self.parents = self.parents[mask_pos]
                self.parents[self.parents > p_pos] -= 1
                # merge labels
                relabel[relabel == p_pos] = relabel[c_pos]
                self.k -= 1
                
                # compute new features
                for fid in self.features.keys():
                    # replace feature
                    # (without the API since self is in an inconsistent state)
                    dj = self.get_feature(fid)
                    dj[c_pos] = np.hstack((dj[self.select_id(c_id)], 
                                           dj[self.select_id(p_id)]))
                    del dj[p_pos]
                    self.features[fid] = dj

                # compute new roi features
                for fid in self.roi_features.keys():
                    dj = self.get_roi_feature(fid)
                    if fid in pull_features:
                        # modify only if `pull` requested
                        dj[c_pos] = dj[p_pos]
                    self.roi_features[fid] = dj[mask_pos]
                                        
        # update HROI structure
        self.label[self.label > -1] = relabel[self.label[self.label > - 1]]
        self.recompute_labels()

    def get_parents(self):
        """Return the parent of each node in the hierarchy

        The parents are represented by their position in the nodes flat list.

        TODO:
        The purpose of this class API is not to rely on this order, so
        we should have self.parents as a list of ids instead of a list of
        positions
        """
        return self.parents

    def get_leaves_id(self):
        """Return the ids of the leaves.

        """
        if self.k == 0:
            return np.array([])
        # locate the positions of the children of each node
        is_leaf_aux = [np.where(self.parents == k)[0] for k in range(self.k)]
        # select nodes that has no child (different from themselves)
        is_leaf = np.asarray(
            [(len(child) == 0) or (len(child) == 1 and child[0] == i)
             for i, child in enumerate(is_leaf_aux)])
        # finaly return ids
        return self.get_id()[is_leaf]

    def reduce_to_leaves(self):
        """Create a  new set of rois which are only the leaves of self.

        Modification of the structure is done in place. One way therefore
        want to work on a copy a of a given HROI oject.

        """
        if self.k == 0:
            # handle the empy HROI case
            return HierarchicalROI(
                self.domain, -np.ones(self.domain.size), np.array([]))
        leaves_id = self.get_leaves_id()
        self.select_roi(leaves_id)

    def copy(self):
        """ Returns a copy of self.

        self.domain is not copied.

        """
        cp = HierarchicalROI(
            self.domain, self.label.copy(), self.parents.copy(), self.get_id())
        # copy features
        for fid in self.features.keys():
            cp.set_feature(fid, self.get_feature(fid))
        # copy ROI features
        for fid in self.roi_features.keys():
            cp.set_roi_feature(fid, self.get_roi_feature(fid))
        return cp

    def representative_feature(self, fid, method='mean', id=None,
                               ignore_children=True, assess_quality=True):
        """Compute a ROI representative of a given feature.

        Parameters
        ----------
        fid: str,
          Feature id
        method: str,
          Method used to compute a representative.
          Chosen among 'mean' (default), 'max', 'median', 'min',
          'weighted mean'.
        id: any hashable type
          Id of the ROI from which we want to extract a representative feature.
          Can be None (default) if we want to get all ROIs's representatives.
        ignore_children: bool,
          Specify if the volume of the node should include
          (ignore_children = False) or not the one of its children
          (ignore_children = True).
        assess_quality: bool
          If True, a new roi feature is created, which represent the quality
          of the feature representative (the number of non-nan value for the
          feature over the ROI size).
          Default is False.

        """
        rf = []
        eps = 1.e-15
        feature_quality = np.zeros(self.k)
        for i, k in enumerate(self.get_id()):
            f = self.get_feature(fid, k)
            p_pos = self.select_id(k)
            if not ignore_children:
                # also include the children features
                desc = np.nonzero(self.parents == p_pos)[0]
                if p_pos in desc:
                    desc = desc[desc != p_pos]
                for c in desc:
                    f = np.concatenate(
                        (f, self.get_feature(fid, self.get_id()[c])))
            # NaN-resistant representative
            if f.ndim == 2:
                nan = np.isnan(f.sum(1))
            else:
                nan = np.isnan(f)
            # feature quality
            feature_quality[i] = (~nan).sum() / float(nan.size)
            # compute representative
            if method == "mean":
                rf.append(np.mean(f[~nan], 0))
            if method == "weighted mean":
                lvk = self.get_local_volume(k)
                if not ignore_children:
                    # append weights for children's voxels
                    for c in desc:
                        lvk = np.concatenate(
                            (lvk,
                             self.get_local_volume(fid, self.select_id(c))))
                tmp = np.dot(lvk[~nan], f[~nan].reshape((-1, 1))) / \
                    np.maximum(eps, np.sum(lvk[~nan]))
                rf.append(tmp)
            if method == "min":
                rf.append(np.min(f[~nan]))
            if method == "max":
                rf.append(np.max(f[~nan]))
            if method == "median":
                rf.append(np.median(f[~nan], 0))
        if id is not None:
            summary_feature = rf[self.select_id(id)]
        else:
            summary_feature = rf

        if assess_quality:
            self.set_roi_feature('%s_quality' % fid, feature_quality)
        return np.array(summary_feature)


def make_hroi_from_subdomain(sub_domain, parents):
    """Instantiate an HROi from a SubDomain instance and parents

    """
    hroi = HierarchicalROI(sub_domain.domain, sub_domain.label, parents)
    # set features
    for fid in sub_domain.features.keys():
        hroi.set_feature(fid, sub_domain.get_feature(fid))
    # set ROI features
    for fid in sub_domain.roi_features.keys():
        hroi.set_roi_feature(fid, sub_domain.get_roi_feature(fid))
    return hroi



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function
__doc__ = """
This examples performs sifferent kinds of (2D and 3D) plots
of a given activation map.

Needs matplotlib.

Author : Bertrand Thirion, 2012
"""
print(__doc__)

from os import path

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

from nibabel import load

from nipy.labs.viz import plot_map, cm

# Local import
from get_data_light import DATA_DIR, get_second_level_dataset

#######################################
# Data and analysis parameters
#######################################

input_image = path.join(DATA_DIR, 'spmT_0029.nii.gz')
if not path.exists(input_image):
    get_second_level_dataset()

brain_map = load(input_image)
vmin, vmax = brain_map.get_data().min(), brain_map.get_data().max()

# make a simple 2D plot
plot_map(brain_map.get_data(), brain_map.get_affine(),
         cmap=cm.cold_hot,
         vmin=vmin,
         vmax=vmax,
         anat=None,
         figure=10,
         threshold=3)

# More plots using 3D
if True:  # replace with False to skip this
    plot_map(brain_map.get_data(), brain_map.get_affine(),
             cmap=cm.cold_hot,
             vmin=vmin,
             vmax=vmax,
             anat=None,
             figure=11,
             threshold=3, do3d=True)

    from nipy.labs import viz3d
    try:
        viz3d.plot_map_3d(brain_map.get_data(), brain_map.get_affine(),
                        cmap=cm.cold_hot,
                        vmin=vmin,
                        vmax=vmax,
                        anat=None,
                        threshold=4)
    except ImportError:
        print("Need mayavi for 3D visualization")

plt.show()



"""
Module for computation of mixed effects statistics with an EM algorithm.
i.e.
solves problems of the form
y = X beta + e1 + e2,
where X and Y are known, e1 and e2 are centered with diagonal covariance.
V1 = var(e1) is known, and V2 = var(e2) = lambda identity.
the code estimates beta and lambda using an EM algorithm.
Likelihood ratio tests can then be used to test the columns of beta.

Author: Bertrand Thirion, 2012.

>>> N, P = 15, 500
>>> V1 = np.random.randn(N, P) ** 2
>>> effects = np.random.randn(P) > 0
>>> Y = generate_data(np.ones(N), effects, .25, V1)
>>> T1 = one_sample_ttest(Y, V1, n_iter=5)
>>> T1 = [T1[effects == x] for x  in np.unique(effects)]
>>> T2 = [t_stat(Y)[effects == x] for x  in np.unique(effects)]
>>> assert np.array([t1.std() < t2.std() for t1, t2 in zip(T1, T2)]).all()
"""

import numpy as np

EPS = 100 * np.finfo(float).eps


def generate_data(X, beta, V2, V1):
    """ Generate a group of individuals from the provided parameters

    Parameters
    ----------
    X: array of shape (n_samples, n_reg),
       the design matrix of the model
    beta: float or array of shape (n_reg, n_tests),
          the associated effects
    V2: float or array of shape (n_tests),
              group variance
    V1: array of shape(n_samples, n_tests),
               the individual variances

    Returns
    -------
    Y: array of shape(n_samples, n_tests)
       the individual data related to the two-level normal model
    """
    # check that the variances are  positive
    if (V1 < 0).any():
        raise ValueError('Variance should be positive')
    Y = np.random.randn(*V1.shape)
    Y *= np.sqrt(V2 + V1)
    if X.ndim == 1:
        X = X[:, np.newaxis]
    if np.isscalar(beta):
        beta = beta * np.ones((X.shape[1], V1.shape[1]))
    if beta.ndim == 1:
        beta = beta[np.newaxis]

    Y += np.dot(X, beta)
    return Y


def check_arrays(Y, V1):
    """Check that the given data can be used for the models

    Parameters
    ----------
    Y: array of shape (n_samples, n_tests) or (n_samples)
       the estimated effects
    V1: array of shape (n_samples, n_tests) or (n_samples)
          first-level variance
    """
    if (V1 < 0).any():
        raise ValueError("a negative variance has been provided")

    if np.size(Y) == Y.shape[0]:
        Y = Y[:, np.newaxis]

    if np.size(V1) == V1.shape[0]:
        V1 = V1[:, np.newaxis]

    if Y.shape != V1.shape:
        raise ValueError("Y and V1 do not have the same shape")
    return Y, V1


def t_stat(Y):
    """ Returns the t stat of the sample on each row of the matrix

    Parameters
    ----------
    Y, array of shape (n_samples, n_tests)

    Returns
    -------
    t_variates, array of shape (n_tests)
    """
    return Y.mean(0) / Y.std(0) * np.sqrt(Y.shape[0] - 1)


class MixedEffectsModel(object):
    """Class to handle multiple one-sample mixed effects models
    """

    def __init__(self, X, n_iter=5, verbose=False):
        """
        Set the effects and first-level variance,
        and initialize related quantities

        Parameters
        ----------
        X: array of shape(n_samples, n_effects),
           the design matrix
        n_iter: int, optional,
               number of iterations of the EM algorithm
        verbose: bool, optional, verbosity mode
        """
        self.n_iter = n_iter
        self.verbose = verbose
        self.X = X
        self.pinv_X = np.linalg.pinv(X)

    def log_like(self, Y, V1):
        """ Compute the log-likelihood of (Y, V1) under the model

        Parameters
        ----------
        Y, array of shape (n_samples, n_tests) or (n_samples)
           the estimated effects
        V1, array of shape (n_samples, n_tests) or (n_samples)
            first-level variance

        Returns
        -------
        logl: array of shape self.n_tests,
              the log-likelihood of the model
        """
        Y, V1 = check_arrays(Y, V1)
        tvar = self.V2 + V1
        logl = np.sum(((Y - self.Y_) ** 2) / tvar, 0)
        logl += np.sum(np.log(tvar), 0)
        logl += np.log(2 * np.pi) * Y.shape[0]
        logl *= (- 0.5)
        return logl

    def predict(self, Y, V1):
        """Return the log_likelihood of the data.See the log_like method"""
        return self.log_like(Y, V1)

    def score(self, Y, V1):
        """Return the log_likelihood of the data. See the log_like method"""
        return self.log_like(Y, V1)

    def _one_step(self, Y, V1):
        """Applies one step of an EM algorithm to estimate self.mean_, self.var

        Parameters
        ----------
        Y, array of shape (n_samples, n_tests) or (n_samples)
              the estimated effects
        V1, array of shape (n_samples, n_tests) or (n_samples)
                 first-level variance
        """
        # E step
        prec = 1. / (self.V2 + V1)
        Y_ = prec * (self.V2 * Y + V1 * self.Y_)
        cvar = V1 * self.V2 * prec

        # M step
        self.beta_ = np.dot(self.pinv_X, Y_)
        self.Y_ = np.dot(self.X, self.beta_)
        self.V2 = np.mean((Y_ - self.Y_) ** 2, 0) + cvar.mean(0)

    def fit(self, Y, V1):
        """ Launches the EM algorithm to estimate self

        Parameters
        ----------
        Y, array of shape (n_samples, n_tests) or (n_samples)
              the estimated effects
        V1, array of shape (n_samples, n_tests) or (n_samples)
                 first-level variance

        Returns
        -------
        self
        """
        # Basic data checks
        if self.X.shape[0] != Y.shape[0]:
            raise ValueError('X and Y must have the same numbers of rows')
        Y, V1 = check_arrays(Y, V1)
        self.beta_ = np.dot(self.pinv_X, Y)
        self.Y_ = np.dot(self.X, self.beta_)
        self.V2 = np.mean((Y - self.Y_) ** 2, 0)

        if self.verbose:
            log_like_init = self.log_like(Y, V1)
            print('Average log-likelihood: ', log_like_init.mean())

        for i in range(self.n_iter):
            self._one_step(Y, V1)

            if self.verbose:
                log_like_ = self.log_like(Y, V1)
                if (log_like_ < (log_like_init - EPS)).any():
                    raise ValueError('The log-likelihood cannot decrease')
                log_like_init = log_like_
                print ('Iteration %d, average log-likelihood: %f' % (
                        i, log_like_.mean()))
        return self


def two_sample_ftest(Y, V1, group, n_iter=5, verbose=False):
    """Returns the mixed effects t-stat for each row of the X
    (one sample test)
    This uses the Formula in Roche et al., NeuroImage 2007

    Parameters
    ----------
    Y: array of shape (n_samples, n_tests)
       the data
    V1: array of shape (n_samples, n_tests)
        first-level variance assocated with the data
    group: array of shape (n_samples)
       a vector of indicators yielding the samples membership
    n_iter: int, optional,
           number of iterations of the EM algorithm
    verbose: bool, optional, verbosity mode

    Returns
    -------
    tstat: array of shape (n_tests),
           statistical values obtained from the likelihood ratio test
    """
    # check that group is correct
    if group.size != Y.shape[0]:
        raise ValueError('The number of labels is not the number of samples')
    if (np.unique(group) != np.array([0, 1])).all():
        raise ValueError('group should be composed only of zeros and ones')

    # create design matrices
    X = np.vstack((np.ones_like(group), group)).T
    return mfx_stat(Y, V1, X, 1, n_iter=n_iter, verbose=verbose,
                    return_t=False, return_f=True)[0]


def two_sample_ttest(Y, V1, group, n_iter=5, verbose=False):
    """Returns the mixed effects t-stat for each row of the X
    (one sample test)
    This uses the Formula in Roche et al., NeuroImage 2007

    Parameters
    ----------
    Y: array of shape (n_samples, n_tests)
       the data
    V1: array of shape (n_samples, n_tests)
        first-level variance assocated with the data
    group: array of shape (n_samples)
       a vector of indicators yielding the samples membership
    n_iter: int, optional,
           number of iterations of the EM algorithm
    verbose: bool, optional, verbosity mode

    Returns
    -------
    tstat: array of shape (n_tests),
           statistical values obtained from the likelihood ratio test
    """
    X = np.vstack((np.ones_like(group), group)).T
    return mfx_stat(Y, V1, X, 1, n_iter=n_iter, verbose=verbose,
                    return_t=True)[0]


def one_sample_ftest(Y, V1, n_iter=5, verbose=False):
    """Returns the mixed effects F-stat for each row of the X
    (one sample test)
    This uses the Formula in Roche et al., NeuroImage 2007

    Parameters
    ----------
    Y: array of shape (n_samples, n_tests)
       the data
    V1: array of shape (n_samples, n_tests)
        first-level variance ssociated with the data
    n_iter: int, optional,
           number of iterations of the EM algorithm
    verbose: bool, optional, verbosity mode

    Returns
    -------
    fstat, array of shape (n_tests),
           statistical values obtained from the likelihood ratio test
    sign, array of shape (n_tests),
          sign of the mean for each test (allow for post-hoc signed tests)
    """
    return mfx_stat(Y, V1, np.ones((Y.shape[0], 1)), 0, n_iter=n_iter,
                    verbose=verbose, return_t=False, return_f=True)[0]


def one_sample_ttest(Y, V1, n_iter=5, verbose=False):
    """Returns the mixed effects t-stat for each row of the X
    (one sample test)
    This uses the Formula in Roche et al., NeuroImage 2007

    Parameters
    ----------
    Y: array of shape (n_samples, n_tests)
       the observations
    V1: array of shape (n_samples, n_tests)
        first-level variance associated with the observations
    n_iter: int, optional,
           number of iterations of the EM algorithm
    verbose: bool, optional, verbosity mode

    Returns
    -------
    tstat: array of shape (n_tests),
           statistical values obtained from the likelihood ratio test
    """
    return mfx_stat(Y, V1, np.ones((Y.shape[0], 1)), 0, n_iter=n_iter,
                    verbose=verbose, return_t=True)[0]


def mfx_stat(Y, V1, X, column, n_iter=5, return_t=True,
             return_f=False, return_effect=False,
             return_var=False, verbose=False):
    """Run a mixed-effects model test on the column of the design matrix

    Parameters
    ----------
    Y: array of shape (n_samples, n_tests)
       the data
    V1: array of shape (n_samples, n_tests)
        first-level variance assocated with the data
    X: array of shape(n_samples, n_regressors)
       the design matrix of the model
    column: int,
            index of the column of X to be tested
    n_iter: int, optional,
           number of iterations of the EM algorithm
    return_t: bool, optional,
              should one return the t test (True by default)
    return_f: bool, optional,
              should one return the F test (False by default)
    return_effect: bool, optional,
              should one return the effect estimate (False by default)
    return_var: bool, optional,
              should one return the variance estimate (False by default)

    verbose: bool, optional, verbosity mode

    Returns
    -------
    (tstat, fstat, effect, var): tuple of arrays of shape (n_tests),
                                 those required by the input return booleans

    """
    # check that X/columns are correct
    column = int(column)
    if X.shape[0] != Y.shape[0]:
        raise ValueError('X.shape[0] is not the number of samples')
    if (column > X.shape[1]):
        raise ValueError('the column index is more than the number of columns')

    # create design matrices
    contrast_mask = 1 - np.eye(X.shape[1])[column]
    X0 = X * contrast_mask

    # instantiate the mixed effects models
    model_0 = MixedEffectsModel(X0, n_iter=n_iter, verbose=verbose).fit(Y, V1)
    model_1 = MixedEffectsModel(X, n_iter=n_iter, verbose=verbose).fit(Y, V1)

    # compute the log-likelihood ratio statistic
    fstat = 2 * (model_1.log_like(Y, V1) - model_0.log_like(Y, V1))
    fstat = np.maximum(0, fstat)
    sign = np.sign(model_1.beta_[column])

    output = ()
    if return_t:
        output += (np.sqrt(fstat) * sign,)
    if return_f:
        output += (fstat,)
    if return_var:
        output += (model_1.V2,)
    if return_effect:
        output += (model_1.beta_[column],)
    return output



"""
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
Test the Bayesian GMM.

fixme : some of these tests take too much time at the moment
to be real unit tests

Author : Bertrand Thirion, 2009
"""

import numpy as np
import numpy.random as nr

from ..bgmm import BGMM, VBGMM, dirichlet_eval, multinomial, dkl_gaussian 

from nose.tools import assert_true

def test_dirichlet_eval():
    # check that the Dirichlet evaluation function sums to one on a simple
    # example
    alpha = np.array([0.5, 0.5])
    sd = 0
    for i in range(10000):
        e = i * 0.0001 + 0.00005
        sd += dirichlet_eval(np.array([e, 1 - e]), alpha)
    assert_true(np.absolute(sd.sum() * 0.0001 - 1) < 0.01)


def test_multinomial():
    """
    test of the generate_multinomial function:
    check that is sums to 1 in a simple case
    """
    n_samples = 100000
    n_classes = 5
    aux = np.reshape(np.random.rand(n_classes), (1, n_classes))
    aux /= aux.sum()
    likelihood = np.repeat(aux, n_samples, 0)
    z = multinomial(likelihood)
    res = np.array([np.sum(z == k) for k in range(n_classes)])
    res = res * 1.0 / n_samples
    assert_true(np.sum((aux-res) ** 2) < 1.e-4)


def test_dkln1():
    dim = 3
    m1 = np.zeros(dim)
    P1 = np.eye(dim)
    m2 = m1
    P2 = P1
    assert_true(dkl_gaussian(m1, P1, m2, P2) == 0)


def test_dkln2():
    dim, offset = 3, 4.
    m1 = np.zeros(dim)
    P1 = np.eye(dim)
    m2 = offset * np.ones(dim)
    P2 = np.eye(dim)
    assert_true(dkl_gaussian(m1, P1, m2, P2) == .5 * dim * offset ** 2)


def test_dkln3():
    dim, scale = 3, 4
    m1, m2 = np.zeros(dim), np.zeros(dim)
    P1, P2 = np.eye(dim), scale * np.eye(dim)
    test1 = .5 * (dim * np.log(scale) + dim * (1. / scale - 1))
    test2 = .5 * (-dim * np.log(scale) + dim * (scale - 1))
    assert_true(dkl_gaussian(m1, P1, m2, P2) == test2)


def test_bgmm_gibbs():
    # Perform the estimation of a gmm using Gibbs sampling
    n_samples, k, dim, niter, offset = 100, 2, 2, 1000, 2.
    x = nr.randn(n_samples,dim)
    x[:30] += offset

    b = BGMM(k,dim)
    b.guess_priors(x)
    b.initialize(x)
    b.sample(x, 1)
    w, cent, prec, pz = b.sample(x, niter, mem=1)
    b.plugin(cent, prec, w)
    z = pz[:, 0]

    # fixme : find a less trivial test
    assert_true(z.max() + 1 == b.k)


def test_gmm_bf(kmax=4, seed=1):
    """ Perform a model selection procedure on a gmm
    with Bayes factor estimations

    Parameters
    ----------
    kmax : range of values that are tested
    seed=False:  int, optionnal
        If seed is not False, the random number generator is initialized
        at a certain value

    fixme : this one often fails. I don't really see why
    """
    n_samples, dim, niter = 30, 2, 1000

    if seed:
        nr = np.random.RandomState([seed])
    else:
        import numpy.random as nr

    x = nr.randn(n_samples, dim)

    bbf = -np.inf
    for k in range(1, kmax):
        b = BGMM(k, dim)
        b.guess_priors(x)
        b.initialize(x)
        b.sample(x, 100)
        w, cent, prec, pz = b.sample(x, niter=niter, mem=1)
        bplugin =  BGMM(k, dim, cent, prec, w)
        bplugin.guess_priors(x)
        bfk = bplugin.bayes_factor(x, pz.astype(np.int))
        if bfk > bbf:
            bestk = k
            bbf = bfk
    assert_true(bestk < 3)


def test_vbgmm():
    """perform the estimation of a variational gmm
    """
    n_samples, dim, offset, k = 100, 2, 2, 2
    x = nr.randn(n_samples, dim)
    x[:30] += offset
    b = VBGMM(k,dim)
    b.guess_priors(x)
    b.initialize(x)
    b.estimate(x)
    z = b.map_label(x)

    # fixme : find a less trivial test
    assert_true(z.max() + 1 == b.k)


def test_vbgmm_select(kmax=6):
    """ perform the estimation of a variational gmm + model selection
    """
    nr.seed([0])
    n_samples, dim, offset=100, 3, 2
    x = nr.randn(n_samples, dim)
    x[:30] += offset
    be = - np.inf
    for  k in range(1, kmax):
        b = VBGMM(k, dim)
        b.guess_priors(x)
        b.initialize(x)
        b.estimate(x)
        ek = b.evidence(x)
        if ek > be:
            be = ek
            bestk = k
    assert_true(bestk < 3)


def test_evidence(k=1):
    """
    Compare the evidence estimated by Chib's method
    with the variational evidence (free energy)
    fixme : this one really takes time
    """
    np.random.seed(0)
    n_samples, dim, offset = 50, 2, 3
    x = nr.randn(n_samples, dim)
    x[:15] += offset

    b = VBGMM(k, dim)
    b.guess_priors(x)
    b.initialize(x)
    b.estimate(x)
    vbe = b.evidence(x)

    niter = 1000
    b = BGMM(k, dim)
    b.guess_priors(x)
    b.initialize(x)
    b.sample(x, 100)
    w, cent, prec, pz = b.sample(x, niter=niter, mem=1)
    bplugin =  BGMM(k, dim, cent, prec, w)
    bplugin.guess_priors(x)
    bfchib = bplugin.bayes_factor(x, pz.astype(np.int), 1)

    assert_true(bfchib > vbe)


if __name__ == '__main__':
    import nose
    nose.run(argv=['', __file__])



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
General utilities for code support.

These are modules that we (broadly-speaking) wrote; packages that other people
wrote, that we ship, go in the nipy.externals tree.
"""

from nibabel.data import make_datasource, DataError, datasource_or_bomber

# Module level datasource instances for convenience
from ..info import DATA_PKGS
templates = datasource_or_bomber(DATA_PKGS['nipy-templates'])
example_data = datasource_or_bomber(DATA_PKGS['nipy-data'])

try:
    example_data.get_filename()
except DataError:
    HAVE_EXAMPLE_DATA = False
else:
    HAVE_EXAMPLE_DATA = True

try:
    templates.get_filename()
except DataError:
    HAVE_TEMPLATES = False
else:
    HAVE_TEMPLATES = True

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



"""
Test the test skip utilities.
"""

import nose
from nipy.utils.skip_test import skip_if_running_nose

# First we must check that during test loading time, our skip_test does
# fire
try:
    skip_if_running_nose()
    skip_test_raised = False
except nose.SkipTest:
    skip_test_raised = True

def test_raise_at_load_time():
    """ Check that SkipTest was raised at load time
    """
    nose.tools.assert_true(skip_test_raised)

def test_not_raise_at_run_time():
    """ Check that SkipTest is not raised at run time 
    """
    try:
        skip_if_running_nose()
    except nose.SkipTest:
        # We need to raise another exception, as nose will capture this
        # one
        raise AssertionError




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
this module contains a function to perform fast distance computation on arrays

Author : Bertrand Thirion, 2008-2011
"""
import numpy as np


def euclidean_distance(X, Y=None):
    """
    Considering the rows of X (and Y=X) as vectors, compute the
    distance matrix between each pair of vectors

    Parameters
    ----------
    X, array of shape (n1,p)
    Y=None, array of shape (n2,p)
            if Y==None, then Y=X is used instead

    Returns
    -------
    ED, array fo shape(n1, n2) with all the pairwise distance
    """
    if Y == None:
        Y = X
    if X.shape[1] != Y.shape[1]:
        raise ValueError("incompatible dimension for X and Y matrices")

    n1 = X.shape[0]
    n2 = Y.shape[0]
    NX = np.reshape(np.sum(X * X, 1), (n1, 1))
    NY = np.reshape(np.sum(Y * Y, 1), (1, n2))
    ED = np.repeat(NX, n2, 1)
    ED += np.repeat(NY, n1, 0)
    ED -= 2 * np.dot(X, Y.T)
    ED = np.maximum(ED, 0)
    ED = np.sqrt(ED)
    return ED



""" Custom doctester based on Numpy doctester

To run doctests via nose, you'll need ``nosetests nipy/testing/doctester.py
--doctest-test``, because this file will be identified as containing tests.
"""
import re
from doctest import register_optionflag
import numpy as np

from ..fixes.numpy.testing.noseclasses import (NumpyDoctest,
                                               NumpyOutputChecker,
                                               NumpyDocTestFinder)

IGNORE_OUTPUT = register_optionflag('IGNORE_OUTPUT')
SYMPY_EQUAL = register_optionflag('SYMPY_EQUAL')
STRIP_ARRAY_REPR = register_optionflag('STRIP_ARRAY_REPR')
IGNORE_DTYPE = register_optionflag('IGNORE_DTYPE')
NOT_EQUAL = register_optionflag('NOT_EQUAL')
FP_4DP =  register_optionflag('FP_4DP')
FP_6DP =  register_optionflag('FP_6DP')

FP_REG = re.compile(r'(?<![0-9a-zA-Z_.])'
                    r'(\d+\.\d+)'
                    r'(e[+-]?\d+)?'
                    r'(?![0-9a-zA-Z_.])')

def round_numbers(in_str, precision):
    """ Replace fp numbers in `in_str` with numbers rounded to `precision`

    Parameters
    ----------
    in_str : str
        string possibly containing floating point numbers
    precision : int
        number of decimal places to round to

    Returns
    -------
    out_str : str
        `in_str` with any floating point numbers replaced with same numbers
        rounded to `precision` decimal places.

    Examples
    --------
    >>> round_numbers('A=0.234, B=12.345', 2)
    'A=0.23, B=12.35'

    Rounds the floating point value as it finds it in the string.  This is even
    true for numbers with exponentials.  Remember that:

    >>> '%.3f' % 0.3339e-10
    '0.000'

    This routine will recognize an exponential as something to process, but only
    works on the decimal part (leaving the exponential part is it is):

    >>> round_numbers('(0.3339e-10, "string")', 3)
    '(0.334e-10, "string")'
    """
    fmt = '%%.%df' % precision
    def dorep(match):
        gs = match.groups()
        res = fmt % float(gs[0])
        if not gs[1] is None:
            res+=gs[1]
        return res
    return FP_REG.sub(dorep, in_str)

ARRAY_REG = re.compile('^\s*array\((.*)\)\s*$', re.DOTALL)
DTYPE_REG = re.compile('\s*,\s+dtype=.*', re.DOTALL)

def strip_array_repr(in_str):
    """ Removes array-specific part of repr from string `in_str`

    This parser only works on lines that contain *only* an array repr (and
    therefore start with ``array``, and end with a close parenthesis.  To remove
    dtypes in array reprs that may be somewhere within the line, use the
    ``IGNORE_DTYPE`` doctest option.

    Parameters
    ----------
    in_str : str
        String maybe containing a repr for an array

    Returns
    -------
    out_str : str
        String from which the array specific parts of the repr have been
        removed.

    Examples
    --------
    >>> arr = np.arange(5, dtype='i2')

    Here's the normal repr:

    >>> arr
    array([0, 1, 2, 3, 4], dtype=int16)

    The repr with the 'array' bits removed:

    >>> strip_array_repr(repr(arr))
    '[0, 1, 2, 3, 4]'
    """
    arr_match = ARRAY_REG.match(in_str)
    if arr_match is None:
        return in_str
    out_str = arr_match.groups()[0]
    return DTYPE_REG.sub('', out_str)

IGNORE_DTYPE_REG = re.compile(',\s+dtype=.*?(?=\))', re.DOTALL)

def ignore_dtype(in_str):
    """ Removes dtype=[dtype] from string `in_str`

    Parameters
    ----------
    in_str : str
        String maybe containing dtype specifier

    Returns
    -------
    out_str : str
        String from which the dtype specifier has been removed.

    Examples
    --------
    >>> arr = np.arange(5, dtype='i2')

    Here's the normal repr:

    >>> arr
    array([0, 1, 2, 3, 4], dtype=int16)

    The repr with the dtype bits removed

    >>> ignore_dtype(repr(arr))
    'array([0, 1, 2, 3, 4])'
    >>> ignore_dtype('something(again, dtype=something)')
    'something(again)'

    Even if there are more closed brackets after the dtype

    >>> ignore_dtype('something(again, dtype=something) (1, 2)')
    'something(again) (1, 2)'

    We need the close brackets to match

    >>> ignore_dtype('again, dtype=something')
    'again, dtype=something'
    """
    return IGNORE_DTYPE_REG.sub('', in_str)


class NipyOutputChecker(NumpyOutputChecker):
    def check_output(self, want, got, optionflags):
        if IGNORE_OUTPUT & optionflags:
            return True
        # When writing tests we sometimes want to assure ourselves that the
        # results are _not_ equal
        wanted_tf = not (NOT_EQUAL & optionflags)
        # Strip dtype
        if IGNORE_DTYPE & optionflags:
            want = ignore_dtype(want)
            got = ignore_dtype(got)
        # Strip array repr from got and want if requested
        if STRIP_ARRAY_REPR & optionflags:
            # STRIP_ARRAY_REPR only matches for a line containing *only* an
            # array repr.  Use IGNORE_DTYPE to ignore a dtype specifier embedded
            # within a more complex line.
            want = strip_array_repr(want)
            got = strip_array_repr(got)
        # If testing floating point, round to required number of digits
        if optionflags & (FP_4DP | FP_6DP):
            if optionflags & FP_4DP:
                dp = 4
            elif optionflags & FP_6DP:
                dp = 6
            want = round_numbers(want, dp)
            got = round_numbers(got, dp)
        # Are the strings equal when run through sympy?
        if SYMPY_EQUAL & optionflags:
            from sympy import sympify
            res = sympify(want) == sympify(got)
            return res == wanted_tf
        # Pass tests through two-pass numpy checker
        res = NumpyOutputChecker.check_output(self, want, got, optionflags)
        # Return True if we wanted True and got True, or if we wanted False and
        # got False
        return res == wanted_tf


class DocTestSkip(object):
    """Object wrapper for doctests to be skipped."""

    def __init__(self,obj):
        self.obj = obj

    def __getattribute__(self,key):
        if key == '__doc__':
            return None
        return getattr(object.__getattribute__(self, 'obj'), key)


class NipyDocTestFinder(NumpyDocTestFinder):

    def _find(self, tests, obj, name, module, source_lines, globs, seen):
        """
        Find tests for the given object and any contained objects, and
        add them to `tests`.

        Add ability to skip doctest
        """
        if hasattr(obj, "skip_doctest") and obj.skip_doctest:
            obj = DocTestSkip(obj)
        NumpyDocTestFinder._find(self,tests, obj, name, module,
                                 source_lines, globs, seen)


class NipyDoctest(NumpyDoctest):
    name = 'nipydoctest'   # call nosetests with --with-nipydoctest
    out_check_class = NipyOutputChecker
    test_finder_class = NipyDocTestFinder

    def set_test_context(self, test):
        # set namespace for tests
        test.globs['np'] = np



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
''' Single subject analysis script for SPM / FIAC '''
import sys
from os.path import join as pjoin
from glob import glob
import numpy as np

from nipy.interfaces.spm import (spm_info, make_job, scans_for_fnames,
                                 run_jobdef, fnames_presuffix, fname_presuffix,
                                 fltcols)


def get_data(data_path, subj_id):
    data_def = {}
    subject_path = pjoin(data_path, 'fiac%s' % subj_id)
    data_def['functionals'] = sorted(
        glob(pjoin(subject_path, 'functional_*.nii')))
    anatomicals = glob(pjoin(subject_path, 'anatomical.nii'))
    if len(anatomicals) == 1:
        data_def['anatomical'] = anatomicals[0]
    elif len(anatomicals) == 0:
        data_def['anatomical'] = None
    else:
        raise ValueError('Too many anatomicals')
    return data_def


def slicetime(data_def):
    sess_scans = scans_for_fnames(data_def['functionals'])
    stinfo = make_job('temporal', 'st', {
            'scans': sess_scans,
            'so':range(1,31,2) + range(2,31,2),
            'tr':2.5,
            'ta':2.407,
            'nslices':float(30),
            'refslice':1
            })
    run_jobdef(stinfo)


def realign(data_def):
    sess_scans = scans_for_fnames(fnames_presuffix(data_def['functionals'], 'a'))
    rinfo = make_job('spatial', 'realign', [{
            'estimate':{
                'data':sess_scans,
                'eoptions':{
                    'quality':0.9,
                    'sep':4.0,
                    'fwhm':5.0,
                    'rtm':True,
                    'interp':2.0,
                    'wrap':[0.0,0.0,0.0],
                    'weight':[]
                    }
                }
            }])
    run_jobdef(rinfo)


def reslice(data_def):
    sess_scans = scans_for_fnames(fnames_presuffix(data_def['functionals'], 'a'))
    rsinfo = make_job('spatial', 'realign', [{
            'write':{
                'data': np.vstack(sess_scans.flat),
                'roptions':{
                    'which':[2, 1],
                    'interp':4.0,
                    'wrap':[0.0,0.0,0.0],
                    'mask':True,
                    }
                }
            }])
    run_jobdef(rsinfo)


def coregister(data_def):
    func1 = data_def['functionals'][0]
    mean_fname = fname_presuffix(func1, 'meana')
    crinfo = make_job('spatial', 'coreg', [{
            'estimate':{
                'ref': [mean_fname],
                'source': [data_def['anatomical']],
                'other': [[]],
                'eoptions':{
                    'cost_fun':'nmi',
                    'sep':[4.0, 2.0],
                    'tol':np.array(
                            [0.02,0.02,0.02,
                             0.001,0.001,0.001,
                             0.01,0.01,0.01,
                             0.001,0.001,0.001]).reshape(1,12),
                    'fwhm':[7.0, 7.0]
                    }
                }
            }])
    run_jobdef(crinfo)


def segnorm(data_def):
    def_tpms = np.zeros((3,1), dtype=np.object)
    spm_path = spm_info.spm_path
    def_tpms[0] = pjoin(spm_path, 'tpm', 'grey.nii'),
    def_tpms[1] = pjoin(spm_path, 'tpm', 'white.nii'),
    def_tpms[2] = pjoin(spm_path, 'tpm', 'csf.nii')
    data = np.zeros((1,), dtype=object)
    data[0] = data_def['anatomical']
    sninfo = make_job('spatial', 'preproc', {
            'data': data,
            'output':{
                'GM':fltcols([0,0,1]),
                'WM':fltcols([0,0,1]),
                'CSF':fltcols([0,0,0]),
                'biascor':1.0,
                'cleanup':False,
                },
            'opts':{
                'tpm':def_tpms,
                'ngaus':fltcols([2,2,2,4]),
                'regtype':'mni',
                'warpreg':1.0,
                'warpco':25.0,
                'biasreg':0.0001,
                'biasfwhm':60.0,
                'samp':3.0,
                'msk':np.array([], dtype=object),
                }
            })
    run_jobdef(sninfo)


def norm_write(data_def):
    sess_scans = scans_for_fnames(fnames_presuffix(data_def['functionals'], 'a'))
    matname = fname_presuffix(data_def['anatomical'],
                              suffix='_seg_sn.mat',
                              use_ext=False)
    subj = {
        'matname': np.zeros((1,), dtype=object),
        'resample': np.vstack(sess_scans.flat),
        }
    subj['matname'][0] = matname
    roptions = {
        'preserve':False,
        'bb':np.array([[-78,-112, -50],[78,76,85.0]]),
        'vox':fltcols([2.0,2.0,2.0]),
        'interp':1.0,
        'wrap':[0.0,0.0,0.0],
        }
    nwinfo = make_job('spatial', 'normalise', [{
            'write':{
                'subj': subj,
                'roptions': roptions,
                }
            }])
    run_jobdef(nwinfo)
    # knock out the list of images, replacing with only one
    subj['resample'] = np.zeros((1,), dtype=object)
    subj['resample'][0] = data_def['anatomical']
    roptions['interp'] = 4.0
    run_jobdef(nwinfo)


def smooth(data_def, fwhm=8.0):
    try:
        len(fwhm)
    except TypeError:
        fwhm = [fwhm] * 3
    fwhm = np.asarray(fwhm, dtype=np.float).reshape(1,3)
    sess_scans = scans_for_fnames(fnames_presuffix(data_def['functionals'], 'wa'))
    sinfo = make_job('spatial', 'smooth',
                     {'data':np.vstack(sess_scans.flat),
                      'fwhm':fwhm,
                      'dtype':0})
    run_jobdef(sinfo)


def process_subject(ddef):
    if not ddef['anatomical']:
        return
    slicetime(ddef)
    realign(ddef)
    reslice(ddef)
    coregister(ddef)
    segnorm(ddef)
    norm_write(ddef)
    smooth(ddef)


def process_subjects(data_path, subj_ids):
    for subj_id in subj_ids:
        ddef = get_data(data_path, subj_id)
        process_subject(ddef)


if __name__ == '__main__':
    try:
        data_path = sys.argv[1]
    except IndexError:
        raise OSError('Need FIAC data path as input')
    try:
        subj_ids = sys.argv[2:]
    except IndexError:
        subj_ids = range(16)
    process_subjects(data_path, subj_ids)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np
import pylab

from nipy.modalities.fmri.utils import events, Symbol, lambdify_t
from nipy.modalities.fmri.hrf import glover

# Symbol for amplitude
a = Symbol('a')

# Some event onsets regularly spaced
onsets = np.linspace(0,50,6)

# Make amplitudes from onset times (greater as function of time)
amplitudes = onsets[:]

# Flip even numbered amplitudes
amplitudes = amplitudes * ([-1, 1] * 3)

# Make event functions
evs = events(onsets, amplitudes=amplitudes, g=a + 0.5 * a**2, f=glover)

# Real valued function for symbolic events
real_evs = lambdify_t(evs)

# Time points at which to sample
t_samples = np.linspace(0,60,601)

pylab.plot(t_samples, real_evs(t_samples), c='r')
for onset, amplitude in zip(onsets, amplitudes):
    pylab.plot([onset, onset],[0, 25 * amplitude], c='b')

pylab.show()



""" Testing models module
"""

import numpy as np

# In fact we're testing methods defined in model
from ..regression import OLSModel

from nose.tools import assert_true, assert_equal, assert_raises
from nose import SkipTest

from numpy.testing import (assert_array_almost_equal,
                           assert_array_equal)


N = 10
X = np.c_[np.linspace(-1,1,N), np.ones((N,))]
Y = np.r_[range(5), range(1,6)]
MODEL = OLSModel(X)
RESULTS = MODEL.fit(Y)

""" R script

::

    X = cbind(0:9 * 2/9 -1, 1)
    Y = as.matrix(c(0:4, 1:5))
    results = lm(Y ~ X-1)
    print(results)
    print(summary(results))

gives::

    Call:
    lm(formula = Y ~ X - 1)

    Coefficients:
    X1     X2
    1.773  2.500

    Residuals:
        Min      1Q  Median      3Q     Max
    -1.6970 -0.6667  0.0000  0.6667  1.6970

    Coefficients:
    Estimate Std. Error t value Pr(>|t|)
    X1   1.7727     0.5455   3.250   0.0117 *
    X2   2.5000     0.3482   7.181 9.42e-05 ***
    ---

    Residual standard error: 1.101 on 8 degrees of freedom
    Multiple R-squared: 0.8859, Adjusted R-squared: 0.8574
    F-statistic: 31.06 on 2 and 8 DF,  p-value: 0.0001694
"""

def test_model():
    # Check basics about the model fit
    # Check we fit the mean
    assert_array_almost_equal(RESULTS.theta[1], np.mean(Y))
    # Check we get the same as R
    assert_array_almost_equal(RESULTS.theta, [1.773, 2.5], 3)
    try:
        percentile = np.percentile
    except AttributeError:
        # Numpy <=1.4.1 does not have percentile function
        raise SkipTest('Numpy does not have percentile function')
    pcts = percentile(RESULTS.resid, [0,25,50,75,100])
    assert_array_almost_equal(pcts, [-1.6970, -0.6667, 0, 0.6667, 1.6970], 4)


def test_t_contrast():
    # Test indivudual t against R
    assert_array_almost_equal(RESULTS.t(0), 3.25)
    assert_array_almost_equal(RESULTS.t(1), 7.181, 3)
    # And contrast
    assert_array_almost_equal(RESULTS.Tcontrast([1,0]).t, 3.25)
    assert_array_almost_equal(RESULTS.Tcontrast([0,1]).t, 7.181, 3)
    # Input matrix checked for size
    assert_raises(ValueError, RESULTS.Tcontrast, [1])
    assert_raises(ValueError, RESULTS.Tcontrast, [1, 0, 0])
    # And shape
    assert_raises(ValueError, RESULTS.Tcontrast, np.array([1, 0])[:,None])


def test_t_output():
    # Check we get required outputs
    exp_t = RESULTS.t(0)
    exp_effect = RESULTS.theta[0]
    exp_sd = exp_effect / exp_t
    res = RESULTS.Tcontrast([1,0])
    assert_array_almost_equal(res.t, exp_t)
    assert_array_almost_equal(res.effect, exp_effect)
    assert_array_almost_equal(res.sd, exp_sd)
    res = RESULTS.Tcontrast([1,0], store=('effect',))
    assert_equal(res.t, None)
    assert_array_almost_equal(res.effect, exp_effect)
    assert_equal(res.sd, None)
    res = RESULTS.Tcontrast([1,0], store=('t',))
    assert_array_almost_equal(res.t, exp_t)
    assert_equal(res.effect, None)
    assert_equal(res.sd, None)
    res = RESULTS.Tcontrast([1,0], store=('sd',))
    assert_equal(res.t, None)
    assert_equal(res.effect, None)
    assert_array_almost_equal(res.sd, exp_sd)
    res = RESULTS.Tcontrast([1,0], store=('effect', 'sd'))
    assert_equal(res.t, None)
    assert_array_almost_equal(res.effect, exp_effect)
    assert_array_almost_equal(res.sd, exp_sd)


def test_f_output():
    # Test f_output
    res = RESULTS.Fcontrast([1,0])
    exp_f = RESULTS.t(0) ** 2
    assert_array_almost_equal(exp_f, res.F)
    # Test arrays work as well as lists
    res = RESULTS.Fcontrast(np.array([1,0]))
    assert_array_almost_equal(exp_f, res.F)
    # Test with matrix against R
    res = RESULTS.Fcontrast(np.eye(2))
    assert_array_almost_equal(31.06, res.F, 2)
    # Input matrix checked for size
    assert_raises(ValueError, RESULTS.Fcontrast, [1])
    assert_raises(ValueError, RESULTS.Fcontrast, [1, 0, 0])
    # And shape
    assert_raises(ValueError, RESULTS.Fcontrast, np.array([1, 0])[:,None])

def test_f_output_new_api():
    res = RESULTS.Fcontrast([1, 0])
    assert_array_almost_equal(res.effect, RESULTS.theta[0])
    assert_array_almost_equal(res.covariance, RESULTS.vcov()[0][0])

def test_conf_int():
    lower_, upper_ = RESULTS.conf_int()
    assert_true((lower_ < upper_).all())
    assert_true((lower_ > upper_ - 10).all())
    lower_, upper_ = RESULTS.conf_int(cols=[1]).T
    assert_true(lower_ < upper_)
    assert_true(lower_ > upper_ - 10)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Build helpers for setup.py

Includes package dependency checks, and code to build the documentation

To build the docs, run::

    python setup.py build_sphinx

This module has to work for python 2 and python 3.
"""
from __future__ import with_statement

# Standard library imports
import sys
import os
from os.path import join as pjoin, dirname, splitext, split as psplit
import zipfile
import warnings
import shutil
from distutils.cmd import Command
from distutils.command.clean import clean
from distutils.command.install_scripts import install_scripts
from distutils.version import LooseVersion
from distutils.dep_util import newer_group
from distutils.errors import DistutilsError

from numpy.distutils.misc_util import appendpath
from numpy.distutils import log
from numpy.distutils.misc_util import is_string

# Python 3 builder
from nisext.py3builder import build_py as old_build_py

# Sphinx import
try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    have_sphinx = False
else:
    have_sphinx = True

# Get project related strings.  Please do not change this line to use
# execfile because execfile is not available in Python 3
_info_fname = pjoin('nipy', 'info.py')
INFO_VARS = {}
exec(open(_info_fname, 'rt').read(), {}, INFO_VARS)

DOC_BUILD_DIR = os.path.join('build', 'html')

CYTHON_MIN_VERSION = INFO_VARS['CYTHON_MIN_VERSION']

################################################################################
# Distutils Command class for installing nipy to a temporary location. 
class TempInstall(Command):
    temp_install_dir = os.path.join('build', 'install')

    def run(self):
        """ build and install nipy in a temporary location. """
        install = self.distribution.get_command_obj('install')
        install.install_scripts = self.temp_install_dir
        install.install_base    = self.temp_install_dir
        install.install_platlib = self.temp_install_dir 
        install.install_purelib = self.temp_install_dir 
        install.install_data    = self.temp_install_dir 
        install.install_lib     = self.temp_install_dir 
        install.install_headers = self.temp_install_dir 
        install.run()

        # Horrible trick to reload nipy with our temporary instal
        for key in sys.modules.keys():
            if key.startswith('nipy'):
                sys.modules.pop(key, None)
        sys.path.append(os.path.abspath(self.temp_install_dir))
        # Pop the cwd
        sys.path.pop(0)
        import nipy

    def initialize_options(self):
        pass
    
    def finalize_options(self):
        pass


################################################################################
# Distutils Command class for API generation 
class APIDocs(TempInstall):
    description = \
    """generate API docs """

    user_options = [
        ('None', None, 'this command has no options'),
        ]


    def run(self):
        # First build the project and install it to a temporary location.
        TempInstall.run(self)
        os.chdir('doc')
        try:
            # We are running the API-building script via an
            # system call, but overriding the import path.
            toolsdir = os.path.abspath(pjoin('..', 'tools'))
            build_templates = pjoin(toolsdir, 'build_modref_templates.py')
            cmd = """%s -c 'import sys; sys.path.append("%s"); sys.path.append("%s"); execfile("%s", dict(__name__="__main__"))'""" \
                % (sys.executable, 
                   toolsdir,
                   self.temp_install_dir,
                   build_templates)
            os.system(cmd)
        finally:
            os.chdir('..')


################################################################################
# Code to copy the sphinx-generated html docs in the distribution.
def relative_path(filename):
    """ Return the relative path to the file, assuming the file is
        in the DOC_BUILD_DIR directory.
    """
    length = len(os.path.abspath(DOC_BUILD_DIR)) + 1
    return os.path.abspath(filename)[length:]



################################################################################
# Distutils Command class to clean
class Clean(clean):

    def run(self):
        clean.run(self)
        api_path = os.path.join('doc', 'api', 'generated')
        if os.path.exists(api_path):
            print("Removing %s" % api_path)
            shutil.rmtree(api_path)
        if os.path.exists(DOC_BUILD_DIR):
            print("Removing %s" % DOC_BUILD_DIR)
            shutil.rmtree(DOC_BUILD_DIR)


################################################################################
# Distutils Command class build the docs
if have_sphinx:
    class MyBuildDoc(BuildDoc):
        """ Sub-class the standard sphinx documentation building system, to
            add logics for API generation and matplotlib's plot directive.
        """

        def run(self):
            self.run_command('api_docs')
            # We need to be in the doc directory for to plot_directive
            # and API generation to work
            os.chdir('doc')
            try:
                BuildDoc.run(self)
            finally:
                os.chdir('..')
            self.zip_docs()

        def zip_docs(self):
            if not os.path.exists(DOC_BUILD_DIR):
                raise OSError('Doc directory does not exist.')
            target_file = os.path.join('doc', 'documentation.zip')
            # ZIP_DEFLATED actually compresses the archive. However, there
            # will be a RuntimeError if zlib is not installed, so we check
            # for it. ZIP_STORED produces an uncompressed zip, but does not
            # require zlib.
            try:
                zf = zipfile.ZipFile(target_file, 'w', 
                                            compression=zipfile.ZIP_DEFLATED)
            except RuntimeError:
                warnings.warn('zlib not installed, storing the docs '
                                'without compression')
                zf = zipfile.ZipFile(target_file, 'w', 
                                            compression=zipfile.ZIP_STORED)    

            for root, dirs, files in os.walk(DOC_BUILD_DIR):
                relative = relative_path(root)
                if not relative.startswith('.doctrees'):
                    for f in files:
                        zf.write(os.path.join(root, f), 
                                os.path.join(relative, 'html_docs', f))
            zf.close()

        def finalize_options(self):
            """ Override the default for the documentation build
                directory.
            """
            self.build_dir = os.path.join(*DOC_BUILD_DIR.split(os.sep)[:-1])
            BuildDoc.finalize_options(self)

else: # failed Sphinx import
    # Raise an error when trying to build docs
    class MyBuildDoc(Command):
        user_options = []
        def run(self):
            raise ImportError(
                "Sphinx is not installed, docs cannot be built")
        def initialize_options(self):
            pass
        def finalize_options(self):
            pass

# Start of numpy.distutils.command.build_py.py copy

class build_py(old_build_py):
    """ Copied verbatim from numpy.distutils.command.build_py.py

    If we are on python 2, this code merely replicates numpy distutils, and we
    could get the same effect by importing build_py from numpy.distutils.  If we
    are on python 3, then `old_build_py` will have python 2to3 routines
    contained, and we also need the methods below to work with numpy distutils.

    If you know a better way to do this, please, send a patch and make me glad.
    """

    def run(self):
        build_src = self.get_finalized_command('build_src')
        if build_src.py_modules_dict and self.packages is None:
            self.packages = build_src.py_modules_dict.keys ()
        old_build_py.run(self)

    def find_package_modules(self, package, package_dir):
        modules = old_build_py.find_package_modules(self, package, package_dir)

        # Find build_src generated *.py files.
        build_src = self.get_finalized_command('build_src')
        modules += build_src.py_modules_dict.get(package,[])

        return modules

    def find_modules(self):
        old_py_modules = self.py_modules[:]
        new_py_modules = filter(is_string, self.py_modules)
        self.py_modules[:] = new_py_modules
        modules = old_build_py.find_modules(self)
        self.py_modules[:] = old_py_modules

        return modules

    # XXX: Fix find_source_files for item in py_modules such that item is 3-tuple
    # and item[2] is source file.

# End of numpy.distutils.command.build_py.py copy

def have_good_cython():
    try:
        from Cython.Compiler.Version import version
    except ImportError:
        return False
    return LooseVersion(version) >= LooseVersion(CYTHON_MIN_VERSION)


def generate_a_pyrex_source(self, base, ext_name, source, extension):
    ''' Monkey patch for numpy build_src.build_src method

    Uses Cython instead of Pyrex.
    '''
    good_cython = have_good_cython()
    if self.inplace or not good_cython:
        target_dir = dirname(base)
    else:
        target_dir = appendpath(self.build_src, dirname(base))
    target_file = pjoin(target_dir, ext_name + '.c')
    depends = [source] + extension.depends
    sources_changed = newer_group(depends, target_file, 'newer') 
    if self.force or sources_changed:
        if good_cython:
            # add distribution (package-wide) include directories, in order
            # to pick up needed .pxd files for cython compilation
            incl_dirs = extension.include_dirs[:]
            dist_incl_dirs = self.distribution.include_dirs
            if not dist_incl_dirs is None:
                incl_dirs += dist_incl_dirs
            import Cython.Compiler.Main
            log.info("cythonc:> %s" % (target_file))
            self.mkpath(target_dir)
            options = Cython.Compiler.Main.CompilationOptions(
                defaults=Cython.Compiler.Main.default_options,
                include_path=incl_dirs,
                output_file=target_file)
            cython_result = Cython.Compiler.Main.compile(source,
                                                       options=options)
            if cython_result.num_errors != 0:
                raise DistutilsError("%d errors while compiling "
                                     "%r with Cython"
                                     % (cython_result.num_errors, source))
        elif sources_changed and os.path.isfile(target_file):
            raise DistutilsError("Cython >=%s required for compiling %r"
                                 " because sources (%s) have changed" %
                                 (CYTHON_MIN_VERSION, source, ','.join(depends)))
        else:
            raise DistutilsError("Cython >=%s required for compiling %r"
                                 " but not available" %
                                 (CYTHON_MIN_VERSION, source))
    return target_file

BAT_TEMPLATE = \
r"""@echo off
REM wrapper to use shebang first line of {FNAME}
set mypath=%~dp0
set pyscript="%mypath%{FNAME}"
set /p line1=<%pyscript%
if "%line1:~0,2%" == "#!" (goto :goodstart)
echo First line of %pyscript% does not start with "#!"
exit /b 1
:goodstart
set py_exe=%line1:~2%
call %py_exe% %pyscript% %*
"""

class install_scripts_nipy(install_scripts):
    """ Make scripts executable on Windows

    Scripts are bare file names without extension on Unix, fitting (for example)
    Debian rules. They identify as python scripts with the usual ``#!`` first
    line. Unix recognizes and uses this first "shebang" line, but Windows does
    not. So, on Windows only we add a ``.bat`` wrapper of name
    ``bare_script_name.bat`` to call ``bare_script_name`` using the python
    interpreter from the #! first line of the script.

    Notes
    -----
    See discussion at
    http://matthew-brett.github.com/pydagogue/installing_scripts.html and
    example at git://github.com/matthew-brett/myscripter.git for more
    background.
    """
    def run(self):
        install_scripts.run(self)
        if not os.name == "nt":
            return
        for filepath in self.get_outputs():
            # If we can find an executable name in the #! top line of the script
            # file, make .bat wrapper for script.
            with open(filepath, 'rt') as fobj:
                first_line = fobj.readline()
            if not (first_line.startswith('#!') and
                    'python' in first_line.lower()):
                log.info("No #!python executable found, skipping .bat "
                            "wrapper")
                continue
            pth, fname = psplit(filepath)
            froot, ext = splitext(fname)
            bat_file = pjoin(pth, froot + '.bat')
            bat_contents = BAT_TEMPLATE.replace('{FNAME}', fname)
            log.info("Making %s wrapper for %s" % (bat_file, filepath))
            if self.dry_run:
                continue
            with open(bat_file, 'wt') as fobj:
                fobj.write(bat_contents)


# The command classes for distutils, used by setup.py
cmdclass = {'api_docs': APIDocs,
            'clean': Clean,
            'build_sphinx': MyBuildDoc,
            'install_scripts': install_scripts_nipy}



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
""" Example of onesample permutation test

Needs *example data* package
"""

import numpy as np

from nipy.labs.group.permutation_test import permutation_test_onesample

from nipy.utils import example_data

# Get group data
group_data = example_data.get_filename('neurospin', 'language_babies',
                                       'offset_002.npz')

f = np.load(group_data)
data, vardata, xyz = f['mat'], f['var'], f['xyz']

# Create one-sample permutation test instance
ptest = permutation_test_onesample(data, xyz, stat_id='wilcoxon')

# Cluster definition: (threshold, diameter)
# Note that a list of definitions can be passed to ptest.calibrate
cluster_def = (ptest.height_threshold(0.01), None)
print(cluster_def)

# Multiple calibration
# To get accurate pvalues, don't pass nperms (default is 1e4)
# Yet it will take longer to run
voxel_res, cluster_res, region_res = ptest.calibrate(nperms=100,
                                                     clusters=[cluster_def])

# Simulated Zmax values for FWER correction
simu_zmax = ptest.zscore(voxel_res['perm_maxT_values'])

# Output regions
## This is a list because several cluster definitions can be accepted
clusters = cluster_res[0]
sizes = clusters['size_values']
clusters_Pcorr = clusters['size_Corr_p_values']

# Simulated cluster sizes
simu_s = clusters['perm_size_values']
simu_smax = clusters['perm_maxsize_values']



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
An implementation of Functions in sympy that allow 'implemented'
functions that can be evaluated when 'lambdified'.

This code is in sympy 0.7.0.  We'll remove this copy in favor of the code in
sympy when - say - 0.7 is in the current and the previous version of Ubuntu.

This module licensed as is sympy, under the BSD license

Sympy docstring:
This module provides convenient functions to transform sympy expressions to
lambda functions which can be used to calculate numerical values very fast.
"""

from __future__ import division
from .compatibility import is_sequence

# There follows a literal copy of sympy.utilities.lambdify
# commit b8aa2de87d537eddde044c13f2eaaab99a5dcfe7
# except where marked by "#NIPY-EDIT"
import inspect

# These are the namespaces the lambda functions will use.
MATH = {}
MPMATH = {}
NUMPY = {}
SYMPY = {}

# Default namespaces, letting us define translations that can't be defined
# by simple variable maps, like I => 1j
# These are separate from the names above because the above names are modified
# throughout this file, whereas these should remain unmodified.
MATH_DEFAULT = {}
MPMATH_DEFAULT = {}
NUMPY_DEFAULT = {"I": 1j}
SYMPY_DEFAULT = {}

# Mappings between sympy and other modules function names.
MATH_TRANSLATIONS = {
    "Abs":"fabs",
    "ceiling":"ceil",
    "E":"e",
    "ln":"log",
}

MPMATH_TRANSLATIONS = {
    "ceiling":"ceil",
    "chebyshevt":"chebyt",
    "chebyshevu":"chebyu",
    "E":"e",
    "I":"j",
    "ln":"log",
    #"lowergamma":"lower_gamma",
    "oo":"inf",
    #"uppergamma":"upper_gamma",
    "LambertW":"lambertw",
    "Matrix":"matrix",
    "conjugate":"conj",
}

NUMPY_TRANSLATIONS = {
    "acos":"arccos",
    "acosh":"arccosh",
    "arg":"angle",
    "asin":"arcsin",
    "asinh":"arcsinh",
    "atan":"arctan",
    "atan2":"arctan2",
    "atanh":"arctanh",
    "ceiling":"ceil",
    "E":"e",
    "im":"imag",
    "ln":"log",
    "Matrix":"matrix",
    "Max":"amax",
    "Min":"amin",
    "oo":"inf",
    "re":"real",
}

# Available modules:
MODULES = {
    "math":(MATH, MATH_DEFAULT, MATH_TRANSLATIONS, ("from math import *",)),
    "mpmath":(MPMATH, MPMATH_DEFAULT, MPMATH_TRANSLATIONS, ("from sympy.mpmath import *",)),
    "numpy":(NUMPY, NUMPY_DEFAULT, NUMPY_TRANSLATIONS, ("from numpy import *",)),
    "sympy":(SYMPY, SYMPY_DEFAULT, {}, ("from sympy.functions import *",
                         "from sympy.matrices import Matrix",
                         "from sympy import Integral, pi, oo, nan, zoo, E, I",)),
}

def _import(module, reload="False"):
    """
    Creates a global translation dictionary for module.

    The argument module has to be one of the following strings: "math",
    "mpmath", "numpy", "sympy".
    These dictionaries map names of python functions to their equivalent in
    other modules.
    """
    # TODO: rewrite this using import_module from sympy.external
    if not module in MODULES:
        raise NameError("This module can't be used for lambdification.")
    namespace, namespace_default, translations, import_commands = MODULES[module]
    # Clear namespace or exit
    if namespace != namespace_default:
        # The namespace was already generated, don't do it again if not forced.
        if reload:
            namespace.clear()
            namespace.update(namespace_default)
        else:
            return

    # It's possible that numpy is not available.
    for import_command in import_commands:
        try:
            exec import_command in {}, namespace
        except ImportError:
            raise ImportError("Can't import %s with command %s" % (module, import_command))

    # Add translated names to namespace
    for sympyname, translation in translations.iteritems():
        namespace[sympyname] = namespace[translation]

def lambdify(args, expr, modules=None, printer=None, use_imps=True):
    """
    Returns a lambda function for fast calculation of numerical values.

    Usage:
    >>> from sympy import sqrt, sin
    >>> from nipy.fixes.sympy.utilities.lambdify import lambdify
    >>> from sympy.abc import x, y, z
    >>> f = lambdify(x, x**2)
    >>> f(2)
    4
    >>> f = lambdify((x,y,z), [z,y,x])
    >>> f(1,2,3)
    [3, 2, 1]
    >>> f = lambdify(x, sqrt(x))
    >>> f(4)
    2.0
    >>> f = lambdify((x,y), sin(x*y)**2)
    >>> f(0, 5)
    0.0

    If not specified differently by the user, Sympy functions are replaced as
    far as possible by either python-math, numpy (if available) or mpmath
    functions - exactly in this order.
    To change this behavior, the "modules" argument can be used.
    It accepts:
     - the strings "math", "mpmath", "numpy", "sympy"
     - any modules (e.g. math)
     - dictionaries that map names of sympy functions to arbitrary functions
     - lists that contain a mix of the arguments above. (Entries that are first
        in the list have higher priority)

    Examples:
    (1) Use one of the provided modules:
        >> f = lambdify(x, sin(x), "math")

        Attention: Functions that are not in the math module will throw a name
                   error when the lambda function is evaluated! So this would
                   be better:
        >> f = lambdify(x, sin(x)*gamma(x), ("math", "mpmath", "sympy"))

    (2) Use some other module:
        >> import numpy
        >> f = lambdify((x,y), tan(x*y), numpy)

        Attention: There are naming differences between numpy and sympy. So if
                   you simply take the numpy module, e.g. sympy.atan will not be
                   translated to numpy.arctan. Use the modified module instead
                   by passing the string "numpy":

        >> f = lambdify((x,y), tan(x*y), "numpy")
        >> f(1, 2)
        -2.18503986326
        >> from numpy import array
        >> f(array([1, 2, 3]), array([2, 3, 5]))
        [-2.18503986 -0.29100619 -0.8559934 ]

    (3) Use own dictionaries:
        >> def my_cool_function(x): ...
        >> dic = {"sin" : my_cool_function}
        >> f = lambdify(x, sin(x), dic)

        Now f would look like:
        >> lambda x: my_cool_function(x)

    Functions present in `expr` can also carry their own numerical
    implementations, in a callable attached to the ``_imp_``
    attribute.  Usually you attach this using the
    ``implemented_function`` factory:

    >>> from sympy.abc import x, y, z
    >>> from nipy.fixes.sympy.utilities.lambdify import lambdify, implemented_function
    >>> from sympy import Function
    >>> f = implemented_function(Function('f'), lambda x : x+1)
    >>> func = lambdify(x, f(x))
    >>> func(4)
    5

    ``lambdify`` always prefers ``_imp_`` implementations to
    implementations in other namespaces, unless the ``use_imps`` input
    parameter is False.
    """
    from sympy.core.symbol import Symbol
    # If the user hasn't specified any modules, use what is available.
    if modules is None:
        # Use either numpy (if available) or python.math where possible.
        # XXX: This leads to different behaviour on different systems and
        #      might be the reason for irreproducible errors.
        try:
            _import("numpy")
            modules = ("math", "numpy", "mpmath", "sympy")
        except ImportError:
            modules = ("math", "mpmath", "sympy")

    # Get the needed namespaces.
    namespaces = []
    # First find any function implementations
    if use_imps:
        namespaces.append(_imp_namespace(expr))
    # Check for dict before iterating
    if isinstance(modules, (dict, str)) or not hasattr(modules, '__iter__'):
        namespaces.append(modules)
    else:
        namespaces += list(modules)
    # fill namespace with first having highest priority
    namespace = {}
    for m in namespaces[::-1]:
        buf = _get_namespace(m)
        namespace.update(buf)

    if hasattr(expr, "atoms"):
        #Try if you can extract symbols from the expression.
        #Move on if expr.atoms in not implemented.
        syms = expr.atoms(Symbol)
        for term in syms:
            namespace.update({str(term): term})

    # Create lambda function.
    lstr = lambdastr(args, expr, printer=printer)

    return eval(lstr, namespace)

def _get_namespace(m):
    """
    This is used by _lambdify to parse its arguments.
    """
    if isinstance(m, str):
        _import(m)
        return MODULES[m][0]
    elif isinstance(m, dict):
        return m
    elif hasattr(m, "__dict__"):
        return m.__dict__
    else:
        raise TypeError("Argument must be either a string, dict or module but it is: %s" % m)

def lambdastr(args, expr, printer=None):
    """
    Returns a string that can be evaluated to a lambda function.

    >>> from sympy.abc import x, y, z
    >>> from nipy.fixes.sympy.utilities.lambdify import lambdastr
    >>> lambdastr(x, x**2)
    'lambda x: (x**2)'
    >>> lambdastr((x,y,z), [z,y,x])
    'lambda x,y,z: ([z, y, x])'

    """
    if printer is not None:
        if inspect.isfunction(printer):
            lambdarepr = printer
        else:
            if inspect.isclass(printer):
                lambdarepr = lambda expr: printer().doprint(expr)
            else:
                lambdarepr = lambda expr: printer.doprint(expr)
    else:
        #XXX: This has to be done here because of circular imports
        from sympy.printing.lambdarepr import lambdarepr

    # Transform everything to strings.
    expr = lambdarepr(expr)
    if isinstance(args, str):
        pass
    elif hasattr(args, "__iter__"):
        args = ",".join(str(a) for a in args)
    else:
        args = str(args)

    return "lambda %s: (%s)" % (args, expr)

def _imp_namespace(expr, namespace=None):
    """ Return namespace dict with function implementations

    We need to search for functions in anything that can be thrown at
    us - that is - anything that could be passed as `expr`.  Examples
    include sympy expressions, as well as tuples, lists and dicts that may
    contain sympy expressions.

    Parameters
    ----------
    expr : object
       Something passed to lambdify, that will generate valid code from
       ``str(expr)``.
    namespace : None or mapping
       Namespace to fill.  None results in new empty dict

    Returns
    -------
    namespace : dict
       dict with keys of implemented function names within `expr` and
       corresponding values being the numerical implementation of
       function

    Examples
    --------
    >>> from sympy.abc import x, y, z
    >>> from nipy.fixes.sympy.utilities.lambdify import implemented_function, _imp_namespace
    >>> from sympy import Function
    >>> f = implemented_function(Function('f'), lambda x : x+1)
    >>> g = implemented_function(Function('g'), lambda x : x*10)
    >>> namespace = _imp_namespace(f(g(x)))
    >>> sorted(namespace.keys())
    ['f', 'g']
    """
    # Delayed import to avoid circular imports
    from sympy.core.function import FunctionClass
    if namespace is None:
        namespace = {}
    # tuples, lists, dicts are valid expressions
    if is_sequence(expr):
        for arg in expr:
            _imp_namespace(arg, namespace)
        return namespace
    elif isinstance(expr, dict):
        for key, val in expr.items():
            # functions can be in dictionary keys
            _imp_namespace(key, namespace)
            _imp_namespace(val, namespace)
        return namespace
    # sympy expressions may be Functions themselves
    func = getattr(expr, 'func', None)
    if isinstance(func, FunctionClass):
        imp = getattr(func, '_imp_', None)
        if not imp is None:
            name = expr.func.__name__
            if name in namespace and namespace[name] != imp:
                raise ValueError('We found more than one '
                                 'implementation with name '
                                 '"%s"' % name)
            namespace[name] = imp
    # and / or they may take Functions as arguments
    if hasattr(expr, 'args'):
        for arg in expr.args:
            _imp_namespace(arg, namespace)
    return namespace

def implemented_function(symfunc, implementation):
    """ Add numerical `implementation` to function `symfunc`

    `symfunc` can by a Function, or a name, in which case we make an
    anonymous function with this name.  The function is anonymous in the
    sense that the name is not unique in the sympy namespace.

    Parameters
    ----------
    symfunc : str or ``sympy.FunctionClass`` instance
       If str, then create new anonymous sympy function with this as
       name.  If `symfunc` is a sympy function, attach implementation to
       function
    implementation : callable
       numerical implementation of function for use in ``lambdify``

    Returns
    -------
    afunc : sympy.FunctionClass instance
       function with attached implementation

    Examples
    --------
    >>> from sympy.abc import x, y, z
    >>> from nipy.fixes.sympy.utilities.lambdify import lambdify, implemented_function
    >>> from sympy import Function
    >>> f = implemented_function(Function('f'), lambda x : x+1)
    >>> lam_f = lambdify(x, f(x))
    >>> lam_f(4)
    5
    """
    #NIPY-EDIT: compatibility between sympy 0.6.x and 0.7.0. UndefinedFunction
    # is, er, not defined in sympy 0.6.x
    try:
        # Delayed import to avoid circular imports
        from sympy.core.function import UndefinedFunction as funcmaker
    except ImportError:
        from sympy import Function, FunctionClass
        funcmaker = lambda name : FunctionClass(Function, name)
    # if name, create anonymous function to hold implementation
    if isinstance(symfunc, basestring):
        symfunc = funcmaker(symfunc)
    #NIPY-EDIT: ends
    # We need to attach as a method because symfunc will be a class
    symfunc._imp_ = staticmethod(implementation)
    return symfunc

# This is a nipy compatability wrapper
import numpy as np

aliased_function = np.deprecate_with_doc(
    'Please use sympy.utilities.implemented_function')(implemented_function)
